unit hiHiLightMemo; {  Memo c   ver 2.51 }

interface

uses Windows,Kol,Share,Debug,WinList,Messages,MMSystem;

const
   MaxLineWidth = 4096;
   msgCaret     = 1;
   msgDblClk    = 2;
   msgShow      = 3;
   msgComplete  = 4;

   {    ,  
           -. }

type 
   {$ifdef F_P}
      TCtrlEx = class;
      PCtrlEx = TCtrlEx;
      TCtrlEx = class(TControl)
   {$else}
      PCtrlEx = ^TCtrlEx;
      TCtrlEx = object(TControl)
   {$endif}
end;

type
   TOptionEdit = ( oeReadOnly, oeOverwrite, oeKeepTrailingSpaces, oeSmartTabs,
                   oeHighlight, oeAutoCompletion );
   TOptionsEdit = set of TOptionEdit;

   TTokenAttrs = packed record
      fontstyle: TFontStyle;
      fontcolor: TColor;
end;
   TOnScanToken = function( Sender: PControl; const FromPos: TPoint;
                            var Attrs: TTokenAttrs ): Integer of object;
   {    :   
     OnScanToken,      FromPos, 
      Lines[ FromPos.Y ],    FromPos.X (  )
        (   Attrs),    .
     ...    ,     . () }

   TFindReplaceOption = ( froBack, froCase, froSpaces, froReplace, froAll );
   TFindReplaceOptions = set of TFindReplaceOption;

   THIhilightMemo = class(THIWinList)
   private
      HE: PCtrlEx;
      Arr:PArray;
      SearchRes:TPoint;
      procedure   SetOption                  (OSet:boolean; Option:TOptionEdit); 
      function    HiLight                    (Sender: PControl;
                                              const FromPos: TPoint;
                                              var Attrs: TTokenAttrs): integer;
      procedure   _OnChange                  (Obj:PObj);
      function    Add                        (const Text:string): integer; override;
      procedure   SetStrings                 (const Value:string); override;
      procedure   _Set                       (var Item:TData; var Val:TData); override;
      function    _Get                       (var Item:TData; var Val:TData): boolean;
      function    _Count:                    integer;
      procedure   _Add                       (var Val:TData);
      procedure   Focused_CaretToView;
   public
      _prop_Overwrite:                       boolean;
      _prop_Strings:                         string;
      _prop_HilightStrings:                  string;
      _prop_AutoCompStrings:                 string;
      _prop_HilightFont:                     TFontRec;
      _prop_AutoComplete:                    boolean;
      _prop_Hilight:                         boolean;
      _prop_MinWordLen:                      integer;
      _prop_FileName:                        string;
      _prop_FindBack:                        boolean;
      _prop_FindCase:                        boolean;
      _prop_FindSpaces:                      boolean;
      _prop_FindReplace:                     boolean;
      _prop_ReplaceAll:                      boolean;
      _prop_SmartTabs:                       boolean;
      _prop_SelectFound:                     boolean;
      _prop_SearchStr:                       string;
      _prop_ReplaceStr:                      string;
      _prop_SearFrom:                        integer;
      _prop_AddType:                         byte;
      _prop_Indent:                          integer;
      _prop_ReadOnly:                        boolean;
      _prop_RightMargin:                     boolean;
      _prop_AutoFocus:                       boolean;
      _prop_ColorRightMargin:                TColor;
      _prop_WidthRightMargin:                integer;
      _prop_ColorUnderLine:                  TColor;

      _data_SearchStr:                       THI_Event;
      _data_ReplaceStr:                      THI_Event;
      _data_SearchFrom:                      THI_Event;
      _data_WordPosition:                    THI_Event;
      _event_onGetBookMIdx:                  THI_Event;
      _event_onClickUnderLineStr:            THI_Event;
 
      destructor Destroy; override;
      procedure Init; override;
      procedure _work_doAdd                  (var _Data:TData; Index:word);
      procedure _work_doClear                (var _Data:TData; Index:word); override;
      procedure _work_doDelete               (var _Data:TData; Index:word);
      procedure _work_doInsert               (var _Data:TData; Index:word);
      procedure _work_doText                 (var _Data:TData; Index:word); 
      procedure _work_doFind                 (var _Data:TData; Index:word);
      procedure _work_doLoad                 (var _Data:TData; Index:word);
      procedure _work_doSave                 (var _Data:TData; Index:word);
      procedure _work_doHilightStrings       (var _Data:TData; Index:word);
      procedure _work_doHilightFont          (var _Data:TData; Index:word);
      procedure _work_doWidthRightMargin     (var _Data:TData; Index:word);
      procedure _work_doSetCaret             (var _Data:TData; Index:word);
      procedure _work_doViewToCaret          (var _Data:TData; Index:word);
      procedure _work_doTopLine              (var _Data:TData; Index:word);
      procedure _work_doLeftCol              (var _Data:TData; Index:word);
      procedure _work_doReplaceSelect        (var _Data:TData; Index:word);
      procedure _work_doDeleteSelect         (var _Data:TData; Index:word);
      procedure _work_doSetSelBegin          (var _Data:TData; Index:word);
      procedure _work_doSetSelEnd            (var _Data:TData; Index:word);
      procedure _work_doSetSelFrom           (var _Data:TData; Index:word);
      procedure _work_doSelectWordUnderCursor(var _Data:TData; Index:word);
      procedure _work_doCut                  (var _Data:TData; Index:word);
      procedure _work_doCopy                 (var _Data:TData; Index:word);
      procedure _work_doPaste                (var _Data:TData; Index:word);
      procedure _work_doUndo                 (var _Data:TData; Index:word);
      procedure _work_doRedo                 (var _Data:TData; Index:word);
      procedure _work_doAutoCompStrings      (var _Data:TData; Index:word);
      procedure _work_doShowAutoComp         (var _Data:TData; Index:word);
      procedure _work_doHideAutoComp         (var _Data:TData; Index:word);
      procedure _work_doClearAutoComp        (var _Data:TData; Index:word);
      procedure _work_doFindReplace          (var _Data:TData; Index:word);
      procedure _work_doFindBack             (var _Data:TData; Index:word);
      procedure _work_doFindCase             (var _Data:TData; Index:word);
      procedure _work_doFindSpaces           (var _Data:TData; Index:word);
      procedure _work_doReplaceAll           (var _Data:TData; Index:word);                
      procedure _work_doSelectFound          (var _Data:TData; Index:word);
      procedure _work_doGetBookMIdx          (var _Data:TData; Index:word);
      procedure _work_doSetBookMIdx          (var _Data:TData; Index:word);
      procedure _work_doColorUnderLine       (var _Data:TData; Index:word);
      procedure _work_doAddType              (var _Data:TData; Index:word);
      procedure _work_doReadOnly             (var _Data:TData; Index:word);
      procedure _work_doAutoComplete         (var _Data:TData; Index:word);
      procedure _work_doOverwrite            (var _Data:TData; Index:word);
      procedure _work_doHilight              (var _Data:TData; Index:word);
      procedure _work_doMinWordLen           (var _Data:TData; Index:word);
      procedure _work_doIndent               (var _Data:TData; Index:word);
      procedure _work_doRightMargin          (var _Data:TData; Index:word);
      procedure _work_doColorRightMargin     (var _Data:TData; Index:word);
      procedure _work_doSmartTabs            (var _Data:TData; Index:word);
      procedure _work_doAutoFocus            (var _Data:TData; Index:word);

      procedure _var_Array                   (var _Data:TData; Index:word);
      procedure _var_Text                    (var _Data:TData; Index:word); override;
      procedure _var_Count                   (var _Data:TData; Index:word);
      procedure _var_TopLine                 (var _Data:TData; Index:word);
      procedure _var_LeftCol                 (var _Data:TData; Index:word);
      procedure _var_LinesPerPage            (var _Data:TData; Index:word);
      procedure _var_LinesVisiblePartial     (var _Data:TData; Index:word);
      procedure _var_ColumnsVisiblePartial   (var _Data:TData; Index:word);
      procedure _var_MaxLineWidthOnPage      (var _Data:TData; Index:word);
      procedure _var_LineHeight              (var _Data:TData; Index:word);
      procedure _var_CharWidth               (var _Data:TData; Index:word);
      procedure _var_SelText                 (var _Data:TData; Index:word);
      procedure _var_CanUndo                 (var _Data:TData; Index:word);
      procedure _var_CanRedo                 (var _Data:TData; Index:word);
      procedure _var_WordAtPos               (var _Data:TData; Index:word);
      procedure _var_WordAtPosMouse          (var _Data:TData; Index:word);
      procedure _var_WordAtPosStartX         (var _Data:TData; Index:word);
      procedure _var_WordAtPosStartY         (var _Data:TData; Index:word);
      procedure _var_PositionX               (var _Data:TData; Index:word);
      procedure _var_PositionY               (var _Data:TData; Index:word);
      procedure _var_PositionMouseX          (var _Data:TData; Index:word);
      procedure _var_PositionMouseY          (var _Data:TData; Index:word);
      procedure _var_SelBeginX               (var _Data:TData; Index:word);
      procedure _var_SelBeginY               (var _Data:TData; Index:word);
      procedure _var_SelEndX                 (var _Data:TData; Index:word);
      procedure _var_SelEndY                 (var _Data:TData; Index:word);
      procedure _var_SelFromX                (var _Data:TData; Index:word);
      procedure _var_SelFromY                (var _Data:TData; Index:word);
      procedure _var_SearchResX              (var _Data:TData; Index:word);
      procedure _var_SearchResY              (var _Data:TData; Index:word);
      procedure _var_SelAvailable            (var _Data:TData; Index:word);
   private
      FCoordinatMouse: TPoint;
      FCaret: TPoint;
      FTopLine: integer;
      FLeftCol: integer;
      FRightMargin: boolean;
      function GetCount: integer;
      function GetText: string;
      function GetLines(Idx: Integer): string;
      procedure SetLines(Idx: Integer; const Value: string);
      function GetSelection: string;
      procedure SetSelection(const Value: String);
      procedure SetSelBegin(const Value: TPoint);
      procedure SetSelEnd(const Value: TPoint);
      procedure SetCaret(const Value: TPoint);
      procedure SetTopLine(const Value: Integer);
      procedure SetLeftCol(const Value: Integer);
      procedure SetText(const Value: String);
      function GetDrawRightMargin: boolean;
      procedure SetDrawRightMargin(const Value: Boolean);
      procedure SetOnScanToken(const Value: TOnScanToken);
   public
      procedure Clear;
      //--------------------------------------------------------------------------
      //      -   
      property Count: integer read GetCount;
      property Lines[Idx: Integer]: string read GetLines write SetLines;
      property Text: string read GetText write SetText;
               { !     Text, Caption 
                  THilightMemo -      
                 .      
                 Edit,  ,   Text ! () }
      procedure InvalidateLine(y: Integer);
      property DrawRightMargin: boolean read GetDrawRightMargin write SetDrawRightMargin;
   public
      RightMarginChars: integer; //   80  . DrawRightMargin
      RightMarginColor: TColor;  //   clBlue
   protected
      FLines: PStrList;// 
      FBufferBitmap: PBitmap;
      procedure DoPaint(DC: HDC);
   protected
      FChangeLevel: Integer;
      procedure Changing; //   
      procedure Changed;  //   
   protected
      FOnScanToken: TOnScanToken;
   public
      //--------------------------------------------------------------------------
      //  
      property OnScanToken: TOnScanToken read FOnScanToken write SetOnScanToken;
   public
      //--------------------------------------------------------------------------
      //  
      Options: TOptionsEdit;
      procedure DeleteLine( y: Integer; UseScroll: Boolean );
      procedure InsertLine( y: Integer; const S: String; UseScroll: Boolean );
      protected
      procedure DeleteLine1( y: Integer );
   public
      //--------------------------------------------------------------------------
      //  
      procedure Indent( delta: Integer );
   public
      //--------------------------------------------------------------------------
      //   ()
      property Caret: TPoint read FCaret write SetCaret;
      function Client2Coord( const P: TPoint ): TPoint;
      procedure CaretToView;
      { CaretToView     ,
                 
         . }
   public
      //--------------------------------------------------------------------------
      //  
   protected
      FSelBegin, FSelEnd, FSelFrom: TPoint; // ,   
                 //  ,     ( )
      procedure SetSel( const Pos1, Pos2, PosFrom: TPoint );
      { SetSel      
            . }
      procedure InvalidateSelection; //     
                                     //  
   public
      property Selection: String read GetSelection write SetSelection;
      property SelBegin: TPoint  read FSelBegin    write SetSelBegin;
      property SelEnd: TPoint    read FSelEnd      write SetSelEnd;
      property SelFrom: TPoint   read FSelFrom     write FSelFrom;
      function SelectionAvailable: Boolean;
      procedure DeleteSelection;
   public
      //--------------------------------------------------------------------------
      //    
   protected
      FMouseDown: Boolean; // TRUE,     
      procedure DoMouseDown( X, Y, Shift: Integer );
      procedure DoMouseMove( X, Y: Integer );
      //--------------------------------------------------------------------------
      //    :
      procedure AdjustHScroll;
      procedure AdjustVScroll;
      procedure DoScroll( Cmd, wParam: DWord );
   public
      //--------------------------------------------------------------------------
      //    (  ,  , 
      //   )
      function WordAtPosStart( const AtPos: TPoint ): TPoint;
      function WordAtPos( const AtPos: TPoint ): String;
      function FindNextWord( const FromPos: TPoint; LookForLettersDigits: Boolean ): TPoint;
      function FindPrevWord( const FromPos: TPoint; LookForLettersDigits: Boolean ): TPoint;
      procedure SelectWordUnderCursor;
      function FindNextTabPos( const FromPos: TPoint ): Integer;
      function FindPrevTabPos( const FromPos: TPoint ): Integer;
   public
      //--------------------------------------------------------------------------
      //  
      procedure DoKeyDown( Key, Shift: Integer );
      procedure DoKeyChar( Key: Char );
   public
      //--------------------------------------------------------------------------
      //      
      property TopLine: Integer read FTopLine write SetTopLine;
      property LeftCol: Integer read FLeftCol write SetLeftCol;
      //--------------------------------------------------------------------------
      //       
      function LinesPerPage: Integer;
      function LinesVisiblePartial: Integer;
      //--------------------------------------------------------------------------
      //        
      function CharWidth: Integer;
      function ColumnsVisiblePartial: Integer;
      function MaxLineWidthOnPage: Integer;
      //--------------------------------------------------------------------------
      //  
      function LineHeight: Integer;
   protected
      FBitmap: PBitmap; //       
      FCharWidth: Integer; //     1- 
      FLineHeight: Integer;//     
        {   . } {    (nesco) }
   public
      //--------------------------------------------------------------------------
      //   
      function CanUndo: Boolean;
      function CanRedo: Boolean;
      procedure Undo;
      procedure Redo;
   protected
      FUndoList, FRedoList: PStrList;
      FUndoingRedoing: Boolean; //  TRUE,    
      procedure DoUndoRedo( List1, List2: PStrList );
   public
      //--------------------------------------------------------------------------
      // 
      procedure SetBookmark( Idx: Integer; const Value: TPoint );
      function GetBookMark( IDx: Integer ): TPoint;
      property Bookmarks[ Idx: Integer ]: TPoint read GetBookMark write SetBookmark;
   protected
      FBookmarks: array[ 0..9 ] of TPoint;
      procedure FixBookmarks( const FromPos: TPoint; deltaX, deltaY: Integer );
   public
      //--------------------------------------------------------------------------
      //  / 
      function FindReplace( S: String; const ReplaceTo: String;
                            const FromPos: TPoint;
                            FindReplaceOptions: TFindReplaceOptions;
                            SelectFound: Boolean ): TPoint;
   public
      //--------------------------------------------------------------------------
      // -
            { !    Applet   
                KOL.TApplet ! () }
      //      .   
      //        .
      AutoCompleteMinWordLength: Integer; //   2 ,  
      procedure AutoAdd2Dictionary( S: PChar ); //     
      procedure AutoCompletionShow;    //  ,  
      procedure AutoCompletionHide;    //   
      procedure AutoCompletionClear;   //  
   protected                           //    -
      FAutoCompletionForm: PControl; //    -
      FAutoCompletionList: PControl; // list view
      FDictionary: PStrListEx; //   ,    
      FAutoFirst, FAutoCount: Integer; //     
      FAutoBestFirst: Integer;         //    -  
      FLastInsertIdx: Integer; //  
      FAutoWord: String; //      -
      FAutoPos: TPoint;  //       
      FnotAdd2Dictionary1time: Boolean; //   1 
      procedure AutoListData( Sender: PControl; Idx, SubItem: Integer;
                              var Txt: String; var ImgIdx: Integer; var State: DWORD;
                              var Store: Boolean );    //    
      function AutoListMessage( var Msg: TMsg; var Rslt: Integer ): Boolean;
      procedure AutoCompleteWord( n: Integer ); //  !
      procedure AutoformCloseEvent( Sender: PObj; var Accept: Boolean );
      procedure AutoFormShowEvent( Sender: PObj );
      procedure FastFillDictionary;   //   Text :=  
  end;

procedure Beep; //  
function CheckKeyMask(State:byte; Code:byte):boolean;

var Upper: array[ Char ] of Char;

procedure InitUpper;

implementation

//  
function IsLetterDigit( C: Char ): Boolean;
begin
   Result:= C in ['A'..'Z','a'..'z',''..'',''..'','','','_', '0'..'9'];
end;

function IsLetter( C: Char ): Boolean;
begin
   Result:= C in ['A'..'Z','a'..'z',''..'',''..'','','','_'];
end;

procedure Beep;
begin
   PlaySound( 'Default', 0, SND_ALIAS );
end;

procedure InitUpper;
var   c: Char;
begin
   for c := #0 to #255 do Upper[ c ] := AnsiUpperCase( c )[ 1 ];
end;

function AnsiStrLComp( S1, S2: PChar; L: Integer ): Integer;
begin
   Result := 0;
   while L > 0 do begin
      Result := Ord( Upper[ S1^ ] ) - Ord( Upper[ S2^ ] );
      if Result <> 0 then Exit;
      Dec( L );
      Inc( S1 );
      Inc( S2 );
   end;
end;

{   .
   -  WM_PAINT  WM_PRINT.
    -        .
    -   - WM_KEYDOWN, WM_CHAR.
    -   .
   . }
function HilightWndFunc( Sender: PControl; var Msg: TMsg; var Rslt: Integer ): Boolean;
var   PaintStruct: TPaintStruct;
      OldPaintDC:  HDC;
      HilightEdit: PCtrlEx;
      HEdit:       THIhilightMemo;
      Cplxity:     Integer;
      CR:          TRect;
      i:           Integer;
      Shift:       Integer; //  WM_KEYDOWN:
      C:           Char;        //  WM_CHAR:
      S:           String;      //  WM_PASTE:

      function GetShiftState: DWORD;
      begin
         Result := 0;
         if GetKeyState( VK_SHIFT ) < 0   then Result := Result or MK_SHIFT;
         if GetKeyState( VK_CONTROL ) < 0 then Result := Result or MK_CONTROL;
      end;
      
begin
   Result:= FALSE; //   -    
   HilightEdit:= PCtrlEx( Sender );
   if HilightEdit = nil then Exit;
   HEdit:= THIhilightMemo( Sender.Tag );
//   if not HilightEdit.Focused then exit;
   CASE Msg.message OF
      WM_PAINT, WM_PRINT: begin
            HilightEdit.fUpdRgn := CreateRectRgn( 0, 0, 0, 0 );
            Cplxity := Integer(
                       GetUpdateRgn( HilightEdit.fHandle, HilightEdit.fUpdRgn, FALSE ) );
            if (Cplxity = NULLREGION) or (Cplxity = ERROR) then begin
               DeleteObject( HilightEdit.fUpdRgn );
               HilightEdit.fUpdRgn := 0;
            end;
            if (HilightEdit.fCollectUpdRgn <> 0) and (HilightEdit.fUpdRgn <> 0) then begin
               if CombineRgn( HilightEdit.fCollectUpdRgn, HilightEdit.fCollectUpdRgn,
                              HilightEdit.fUpdRgn, RGN_OR ) = COMPLEXREGION then begin
                  windows.GetClientRect( HilightEdit.fHandle, CR );
                  DeleteObject( HilightEdit.fCollectUpdRgn );
                  HilightEdit.fCollectUpdRgn := CreateRectRgnIndirect( CR );
               end;
               InvalidateRgn( HilightEdit.fHandle, HilightEdit.fCollectUpdRgn, FALSE{fEraseUpdRgn} );
            end;
            OldPaintDC := HilightEdit.fPaintDC;
            HilightEdit.fPaintDC := Msg.wParam;
            if HilightEdit.fPaintDC = 0 then
               HilightEdit.fPaintDC := BeginPaint( HilightEdit.fHandle, PaintStruct );
            if HilightEdit.fCollectUpdRgn <> 0 then
               SelectClipRgn( HilightEdit.fPaintDC, HilightEdit.fCollectUpdRgn );
            HEdit.DoPaint( HilightEdit.fPaintDC );
            if assigned( HilightEdit.fCanvas ) then
               HilightEdit.fCanvas.Handle := 0;
            if Msg.wParam = 0 then
               EndPaint( HilightEdit.fHandle, PaintStruct );
            HilightEdit.fPaintDC := OldPaintDC;
            if HilightEdit.fUpdRgn <> 0 then
               DeleteObject( HilightEdit.fUpdRgn );
            HilightEdit.fUpdRgn := 0;
            Rslt := 0;
            Result := True;
         end;
      WM_SETFOCUS, WM_KILLFOCUS:
            HilightEdit.Perform(WM_USER + msgCaret,0,0);
      WM_USER + msgCaret:
            HEdit.Caret := HEdit.Caret;
      WM_KEYDOWN: begin
            Shift:= GetShiftState;
            if Assigned( HilightEdit.OnKeyDown ) then
               HilightEdit.OnKeyDown( HilightEdit, Msg.wParam, Shift );
            if Msg.wParam <> 0 then
               HEdit.DoKeyDown( Msg.wParam, Shift );
         end;
      WM_LBUTTONDOWN: begin
            HilightEdit.Focused := TRUE;
            HEdit.DoMouseDown(
               SmallInt( Loword( Msg.lParam ) ),
               SmallInt( hiWord( Msg.lParam ) ),
               integer(CheckKeyMask(1,VK_SHIFT)) );
         end;
      WM_RBUTTONDOWN, WM_MBUTTONDOWN:
            HilightEdit.Focused := TRUE;
      WM_LBUTTONUP: begin
            HEdit.DoMouseMove(
               SmallInt( Loword( Msg.lParam ) ),
               SmallInt( hiWord( Msg.lParam ) ) );
            HEdit.FMouseDown := FALSE;
            ReleaseCapture;
         end;
      WM_MOUSEMOVE: begin
            HEdit.DoMouseMove(
               SmallInt( Loword( Msg.lParam ) ),
               SmallInt( hiWord( Msg.lParam ) ) );
         end;
      WM_MOUSEWHEEL: begin
            i := SmallInt( HiWord( Msg.wParam ) );
            if i div 120 <> 0 then HEdit.TopLine := HEdit.TopLine - i div 120;
         end;
      WM_LBUTTONDBLCLK: begin
            Shift := integer(CheckKeyMask(1,VK_SHIFT));
            HEdit.DoMouseDown(
               SmallInt( Loword( Msg.lParam ) ),
               SmallInt( hiWord( Msg.lParam ) ),
               Shift );
            HEdit.FMouseDown := FALSE;
            ReleaseCapture;
            HilightEdit.Perform(WM_USER+msgDblClk,0,0);
         end;
      WM_USER + msgDblClk: begin
            Shift := Msg.wParam;
            if Shift and MK_SHIFT = 0 then HEdit.SelectWordUnderCursor;
         end;
      WM_CHAR: begin
            C := Char( Msg.wParam );
            if Assigned( HilightEdit.OnChar ) then
               HilightEdit.OnChar( HilightEdit, C, integer(CheckKeyMask(1,VK_SHIFT)) );
            if C <> #0 then
               HEdit.DoKeyChar( C );
         end;
      WM_SIZE: begin
            HEdit.AdjustHScroll;
            HEdit.AdjustVScroll;
         end;
      WM_HScroll:
            HEdit.DoScroll( SC_HSCROLL, Msg.wParam );
      WM_VScroll:
            HEdit.DoScroll( SC_VSCROLL, Msg.wParam );
      WM_SETFONT: begin
            HEdit.FCharWidth := 0;
            HEdit.FLineHeight := 0;
            HEdit.Caret := HEdit.Caret;
         end;
      WM_COPY:
            if HEdit.SelectionAvailable then Text2Clipboard( HEdit.Selection );
      WM_CUT:
            if HEdit.SelectionAvailable then begin
               Text2Clipboard( HEdit.Selection );
               HEdit.DeleteSelection;
            end;
      WM_PASTE: begin
            S := Clipboard2Text;
            if S = '' then Beep
            else HEdit.Selection := S;
         end;
      end;
end;

procedure THIHiLightMemo.Init;
begin
   Control := _NewControl( FParent, 'HILIGHTEDIT',
             WS_CHILD or WS_VISIBLE
             or WS_TABSTOP // WS_TABSTOP     
             or WS_HSCROLL or WS_VSCROLL or WS_BORDER,
             TRUE, nil );
   HE:= PCtrlEx( Control );
   Control.ExStyle := Control.ExStyle or WS_EX_CLIENTEDGE;
   Control.ClsStyle := (Control.ClsStyle or CS_DBLCLKS)
                   and not (CS_HREDRAW or CS_VREDRAW);
   Control.Cursor := LoadCursor( 0, IDC_IBEAM );
   {    }
   Control.Tag:= LongInt(Self);
   Control.AttachProc( HilightWndFunc );
   {   }
   HE.fLookTabKeys := [  ];
   FLines := NewStrList;
   FBitmap := NewDibBitmap( 1, 1, pf32bit );
   FUndoList := NewStrList;
   FRedoList := NewStrList;
   FDictionary := NewStrListEx;
   AutoCompleteMinWordLength := 2;
inherited;
   if _prop_Strings <> '' then Text:=_prop_Strings else Text:=' '; 
   Options := [oeKeepTrailingSpaces];
   if _prop_SmartTabs    then Options := Options + [oeSmartTabs];
   if _prop_Overwrite    then Options := Options + [oeOverwrite];
   if _prop_Hilight      then Options := Options + [oeHighLight];
   if _prop_ReadOnly     then Options := Options + [oeReadOnly];
   if _prop_AutoComplete then begin 
      Options := Options + [oeAutoCompletion];
      AutoCompleteMinWordLength:=_prop_MinWordLen;
      AutoAdd2Dictionary(PChar(_prop_AutoCompStrings));
   end;
   DrawRightMargin := _prop_RightMargin;
   RightMarginChars := _prop_WidthRightMargin;
   RightMarginColor := _prop_ColorRightMargin;
   OnScanToken := Hilight;
   Control.OnChange := _OnChange;
end;

procedure THIHiLightMemo.SetOption;
var   l:TOptionsEdit;
begin
   l:= Options;
   if OSet then include(l,Option)
   else exclude(l,Option);
   Options:= l;
   Focused_CaretToView;
   Control.Invalidate;
end;

destructor THIHiLightMemo.Destroy;
begin
   FLines.free; //  
   FBitmap.free;
   free_and_nil(FBufferBitmap);
   FUndoList.free;
   FRedoList.free;
   if Assigned(FAutoCompletionForm) then begin
      FAutoCompletionForm.Close;
      FAutoCompletionForm:= nil;
   end;
   free_and_nil(FDictionary);
   if Arr <> nil then Dispose(Arr);
   inherited;
end;

function THIHiLightMemo.HiLight;
var   S, str, CS, U : string;
      i, j          : integer;
      HS            : PStrList;
      _Color        : TColor;
begin
   HS:=NewStrList;
TRY
   HS.text := _prop_HilightStrings;
   S := Lines[FromPos.Y];
   Result := 0;
   if S = '' then exit;
   i := FromPos.X;
   Attrs.FontStyle := [ ];
   Attrs.fontcolor := Control.Font.Color;
   for j := 0 to HS.Count - 1 do begin
     U := HS.Items[ j ] + '=';
     str := getTok(U,'=');
     if U <> '' then CS := getTok(U,'=') else CS := ''; 
     _Color := _prop_HilightFont.Color;
     if ((AnsiCompareStrNoCase(Copy( S, i+1, Length(str)), str) = 0)) then begin
        if (CS <> '') then begin
           CS:= UpperCase(CS);
           if (CS[1] in ['0'..'9']) then _Color:= str2int(CS)
           else if CS = 'BLACK'     then _Color:= clBlack
           else if CS = 'MAROON'    then _Color:= clMaroon
           else if CS = 'GREEN'     then _Color:= clGreen
           else if CS = 'OLIVE'     then _Color:= clOlive
           else if CS = 'NAVY'      then _Color:= clNavy
           else if CS = 'PURPLE'    then _Color:= clPurple
           else if CS = 'TEAL'      then _Color:= clTeal
           else if CS = 'GRAY'      then _Color:= clGray
           else if CS = 'SILVER'    then _Color:= clSilver
           else if CS = 'RED'       then _Color:= clRed
           else if CS = 'LIME'      then _Color:= clLime
           else if CS = 'YELLOW'    then _Color:= clYellow
           else if CS = 'BLUE'      then _Color:= clBlue
           else if CS = 'FUCHSIA'   then _Color:= clFuchsia
           else if CS = 'AQUA'      then _Color:= clAqua
           else if CS = 'WHITE'     then _Color:= clWhite
        end;
        Attrs.fontcolor := _Color;
        if U <> '' then deleteTail(U,1);
        U:= UpperCase(U);
        if (U = 'U') and (oeReadOnly in Options) then
           Attrs.fontstyle := TFontStyle({$ifdef F_P}integer(_prop_HilightFont.Style)
                                         {$else}_prop_HilightFont.Style
                                         {$endif})  + [fsUnderline]
        else           
           Attrs.fontstyle := TFontStyle({$ifdef F_P}integer(_prop_HilightFont.Style)
                                         {$else}_prop_HilightFont.Style{$endif});
      end;
   end;
   if S[ i+1 ] <= ' ' then
      while (i < Length(S)) and (S[ i+1 ] <= ' ') do inc(i)
   else
      while (i < Length(S)) and (S[ i+1 ] > ' ') do inc(i);
   Result:= i - FromPos.x;
FINALLY
  HS.free;
END;
end;

procedure THIHiLightMemo.Focused_CaretToView;
begin
   AdjustVScroll;
   AdjustHScroll;
   if not _prop_AutoFocus then exit;
   Control.Focused := True;
   CaretToView;  
end;

procedure THIHiLightMemo.Add;
begin
   Result := Control.Count - 1;
   if oeReadOnly in Options then exit;
   if (Control.Count > 0 )and (Control.Items[Control.Count - 1] <> '') then
   Control.Add(#13#10);
   Result := Control.Add(Text);
end;

procedure THIHiLightMemo.SetStrings;
begin
   Control.Perform(WM_SETTEXT, 0, Longint(@Value[1]));
end;

procedure THIHiLightMemo._OnChange;
begin
   _hi_OnEvent(_event_onChange, Text);
end;

procedure THIHiLightMemo._work_doAdd;
begin
TRY
   if oeReadOnly in Options then exit;
   if _prop_AddType = 0 then
      InsertLine(Count,ReadString(_Data,_data_Str,''),True)
   else
      InsertLine(0,ReadString(_Data,_data_Str,''),True);
FINALLY
   Focused_CaretToView; 
END;
end;

procedure THIHiLightMemo._work_doInsert;
begin
TRY
   if oeReadOnly in Options then exit;
   InsertLine(ToIntIndex(_Data),ToStringEvent(_data_Str),True);
FINALLY
   Focused_CaretToView; 
END;
end;

procedure THIHiLightMemo._work_doClear;
begin
   Text := '';
   _hi_OnEvent( _event_onChange );
   Focused_CaretToView; 
end;

procedure THIHiLightMemo._work_doDelete;
begin
TRY
   if oeReadOnly in Options then exit;
   DeleteLine( ToIntIndex( _Data ), True );
FINALLY
   Focused_CaretToView; 
END;  
end;

procedure THIHiLightMemo._work_doText;
begin
   Text := ToString(_Data);
   _hi_OnEvent( _event_onChange );
   Focused_CaretToView;
end;

procedure THIHiLightMemo._work_doLoad;
var   fn:string;
      Strm:PStream;
      SL:PStrList;
begin
   fn := ReadString(_Data,_data_FileName,_prop_FileName);
   if not FileExists(fn) then exit;
   SL := NewStrList;
   Strm := NewReadFileStream(fn);
   Strm.Position := 0;
   SL.LoadFromStream(Strm, false);
   Text := SL.Text;
   free_and_nil(Strm);
   SL.free;
   Focused_CaretToView;   
   _hi_OnEvent(_event_onChange);   
end;

procedure THIHiLightMemo._work_doSave;
var   fn:string;
      Strm:PStream;
      SL:PStrList;
begin
   SL := NewStrList;
   SL.Text := Text;
   fn := ReadString(_Data,_data_FileName,_prop_FileName);
   Strm := NewWriteFileStream(fn);
   Strm.Position := 0;
   SL.SaveToStream(Strm);
   free_and_nil(Strm);
   Sl.free;
   Focused_CaretToView;  
end; 

procedure THIHiLightMemo._work_doCut;
begin
TRY
   if oeReadOnly in Options then exit;
   Control.Perform( WM_CUT, 0, 0 );
FINALLY
   Focused_CaretToView;
END;
end; 

procedure THIHiLightMemo._work_doCopy;
begin
   Control.Perform( WM_COPY, 0, 0 );
   Focused_CaretToView;
end; 

procedure THIHiLightMemo._work_doPaste;
begin
TRY
   if oeReadOnly in Options then exit;
   Control.Perform( WM_PASTE, 0, 0 );
FINALLY
   Focused_CaretToView;
END;
end; 

procedure THIHiLightMemo._work_doFind;
var   FROptions:TFindReplaceOptions;
begin
   FROptions := [ ]; 
   if _prop_ReplaceAll  then FROptions := FROptions + [froAll];
   if _prop_FindBack    then FROptions := FROptions + [froBack];
   if _prop_FindCase    then FROptions := FROptions + [froCase];
   if _prop_FindSpaces  then FROptions := FROptions + [froSpaces];
   if _prop_FindReplace then FROptions := FROptions + [froReplace];
   SearchRes := FindReplace(ReadString(_Data,_data_SearchStr,_prop_SearchStr),
                ReadString(_Data,_data_ReplaceStr,_prop_ReplaceStr),
                MakePoint(ToIntegerEvent(_data_SearchFrom) and $ffff,
                ToIntegerEvent(_data_SearchFrom) shr 16),FROptions,_prop_SelectFound);
   Focused_CaretToView;
end;

procedure THIHiLightMemo._work_doHilightStrings;
begin
   _prop_HiLightStrings := ToString(_Data);
   Focused_CaretToView;   
   Control.Invalidate;
end;

procedure THIHiLightMemo._work_doHilightFont;
begin
   if _IsFont(_Data) then
      with pfontrec(_data.idata)^ do begin
         _prop_HilightFont.Style := Style;
         _prop_HilightFont.Color := Color;
      end;
   Focused_CaretToView;   
end;

procedure THIHiLightMemo._work_doWidthRightMargin;
begin
   RightMarginChars := ToInteger(_Data);
   Focused_CaretToView;  
   Control.Invalidate;
end;

procedure THIHiLightMemo._work_doSetCaret;
var   pos:cardinal;
      pnt:TPoint;
begin
   pos := Cardinal(ToInteger(_Data));
   pnt.X := pos and $ffff;
   pnt.Y := pos shr 16;
   Caret := pnt;
   Focused_CaretToView;   
end;

procedure THIHiLightMemo._work_doReplaceSelect;
begin
   Selection := ToString(_Data);
   Focused_CaretToView;   
end;

procedure THIHiLightMemo._work_doDeleteSelect;
begin
   DeleteSelection;
   Focused_CaretToView;   
end;

procedure THIHiLightMemo._work_doSelectWordUnderCursor;
begin
   SelectWordUnderCursor;
   Focused_CaretToView;
end;   

procedure THIHiLightMemo._work_doSetSelBegin;
var   pos:cardinal;
      pnt:TPoint;
begin
   pos := Cardinal(ToInteger(_Data));
   pnt.X := pos and $ffff;
   pnt.Y := pos shr 16;
   SelBegin := pnt;
   Focused_CaretToView;
end;

procedure THIHiLightMemo._work_doSetSelEnd;
var   pos:cardinal;
      pnt:TPoint;
begin
   pos := Cardinal(ToInteger(_Data));
   pnt.X := pos and $ffff;
   pnt.Y := pos shr 16;
   SelEnd := pnt;
   Focused_CaretToView;
end;

procedure THIHiLightMemo._work_doSetSelFrom;
var   pos:cardinal;
      pnt:TPoint;
begin
   pos := Cardinal(ToInteger(_Data));
   pnt.X := pos and $ffff;
   pnt.Y := pos shr 16;
   SelFrom := pnt;
   Focused_CaretToView;
end;

procedure THIHiLightMemo._work_doGetBookMIdx;
var   idx:integer;
begin
   idx := ToIntIndex(_Data);
   if (idx < 0) or (idx > 9) then exit; 
   _hi_onEvent(_event_onGetBookMIdx, Bookmarks[idx].Y shl 16 + Bookmarks[idx].X);
   Focused_CaretToView;
end;

procedure THIHiLightMemo._work_doSetBookMIdx;
var   idx:integer;
begin
   idx := ToIntIndex(_Data);
   if (idx < 0) or (idx > 9) then exit; 
   Bookmarks[idx] := Caret;
   Focused_CaretToView;
end;

procedure THIHiLightMemo._Set;
var   ind:integer;
begin
   ind:= ToIntIndex(Item);
   if (ind >= 0) and (ind < Count) and not (oeReadOnly in Options) then
      Lines[ind] := ToString(Val);
   inherited;
end;

function THIHiLightMemo._Get;
var   ind:integer;
begin
   ind:= ToIntIndex(Item);
   if (ind >= 0) and (ind < Count) then begin
      Result:= true;
      dtString(Val, Lines[ind]);
   end
   else Result:= false;
end;

procedure THIHiLightMemo._Add;
begin
   if oeReadOnly in Options then exit;
   InsertLine(Count, ToString(Val), false);
end;

function  THIHiLightMemo._Count;
begin
   Result:= Count;
end;

procedure THIHiLightMemo._var_Array;
begin
   if Arr = nil then Arr:= CreateArray(_Set, _Get, _Count, _Add);
   dtArray(_Data, Arr);
end;

procedure THIHiLightMemo._work_doUndo;
begin
TRY
   if oeReadOnly in Options then exit;
   Undo;
FINALLY
   Focused_CaretToView; 
END;
end;

procedure THIHiLightMemo._work_doRedo;
begin
TRY
   if oeReadOnly in Options then exit;
   Redo;
FINALLY
   Focused_CaretToView; 
END;
end;

procedure THIHiLightMemo._work_doViewToCaret;       begin Focused_CaretToView;                       end;
procedure THIHiLightMemo._work_doTopLine;           begin TopLine := ToInteger(_Data);               end;
procedure THIHiLightMemo._work_doLeftCol;           begin LeftCol := ToInteger(_Data);               end;
procedure THIHiLightMemo._work_doAutoCompStrings;   begin AutoAdd2Dictionary(PChar(ToString(_Data)));end;
procedure THIHiLightMemo._work_doShowAutoComp;      begin AutoCompletionShow;                        end;
procedure THIHiLightMemo._work_doHideAutoComp;      begin AutoCompletionHide;                        end;
procedure THIHiLightMemo._work_doClearAutoComp;     begin AutoCompletionClear;                       end;
procedure THIHiLightMemo._var_Text;                 begin dtString(_Data, Text);                     end;
procedure THIHiLightMemo._var_Count;                begin dtInteger(_Data, Count);                   end;
procedure THIHiLightMemo._var_SelText;              begin dtString(_Data, Selection);                end;
procedure THIHiLightMemo._var_TopLine;              begin dtInteger(_Data, TopLine);                 end;
procedure THIHiLightMemo._var_LeftCol;              begin dtInteger(_Data, LeftCol);                 end;
procedure THIHiLightMemo._var_LinesPerPage;         begin dtInteger(_Data, LinesPerPage);            end;
procedure THIHiLightMemo._var_LineHeight;           begin dtInteger(_Data, LineHeight);              end;
procedure THIHiLightMemo._var_CharWidth;            begin dtInteger(_Data, CharWidth);               end;
procedure THIHiLightMemo._var_SelAvailable;         begin dtInteger(_Data,Byte(SelectionAvailable)); end;
procedure THIHiLightMemo._var_CanUndo;              begin dtInteger(_Data,Byte(CanUndo));            end;
procedure THIHiLightMemo._var_CanRedo;              begin dtInteger(_Data,Byte(CanRedo));            end;
procedure THIHiLightMemo._var_PositionX;            begin dtInteger(_Data, Caret.X);                 end;
procedure THIHiLightMemo._var_PositionY;            begin dtInteger(_Data, Caret.Y);                 end;
procedure THIHiLightMemo._var_PositionMouseX;       begin dtInteger(_Data, FCoordinatMouse.X);       end;
procedure THIHiLightMemo._var_PositionMouseY;       begin dtInteger(_Data, FCoordinatMouse.Y);       end;
procedure THIHiLightMemo._var_SelBeginX;            begin dtInteger(_Data, SelBegin.X);              end;
procedure THIHiLightMemo._var_SelBeginY;            begin dtInteger(_Data, SelBegin.Y);              end;
procedure THIHiLightMemo._var_SelEndX;              begin dtInteger(_Data, SelEnd.X);                end;
procedure THIHiLightMemo._var_SelEndY;              begin dtInteger(_Data, SelEnd.Y);                end;
procedure THIHiLightMemo._var_SelFromX;             begin dtInteger(_Data, SelFrom.X);               end;
procedure THIHiLightMemo._var_SelFromY;             begin dtInteger(_Data, SelFrom.Y);               end;
procedure THIHiLightMemo._var_SearchResX;           begin dtInteger(_Data, SearchRes.X);             end;
procedure THIHiLightMemo._var_SearchResY;           begin dtInteger(_Data, SearchRes.Y);             end;
procedure THIHiLightMemo._var_LinesVisiblePartial;  begin dtInteger(_Data, LinesVisiblePartial);     end;
procedure THIHiLightMemo._var_ColumnsVisiblePartial;begin dtInteger(_Data, ColumnsVisiblePartial);   end;
procedure THIHiLightMemo._var_MaxLineWidthOnPage;   begin dtInteger(_Data, MaxLineWidthOnPage);      end;
procedure THIHiLightMemo._work_doFindReplace;       begin _prop_FindReplace := ReadBool(_Data);      end;
procedure THIHiLightMemo._work_doFindBack;          begin _prop_FindBack := ReadBool(_Data);         end;
procedure THIHiLightMemo._work_doFindCase;          begin _prop_FindCase := ReadBool(_Data);         end;
procedure THIHiLightMemo._work_doFindSpaces;        begin _prop_FindSpaces := ReadBool(_Data);       end;
procedure THIHiLightMemo._work_doReplaceAll;        begin _prop_ReplaceAll := ReadBool(_Data);       end;    
procedure THIHiLightMemo._work_doSelectFound;       begin _prop_SelectFound := ReadBool(_Data);      end; 
procedure THIHiLightMemo._work_doAddType;           begin _prop_AddType := ToInteger(_Data);         end;
procedure THIHiLightMemo._work_doAutoFocus;         begin _prop_AutoFocus := ReadBool(_Data);        end;

procedure THIHiLightMemo._work_doColorUnderLine;
begin
   _prop_ColorUnderLine := ToInteger(_Data);
   Control.Invalidate;
   Focused_CaretToView;   
end;

procedure THIHiLightMemo._work_doReadOnly;
begin
   _prop_ReadOnly:= ReadBool(_Data);
   SetOption(_prop_ReadOnly, oeReadOnly);
end;

procedure THIHiLightMemo._work_doSmartTabs;
begin
   _prop_SmartTabs:= ReadBool(_Data);
   SetOption(_prop_SmartTabs, oeSmartTabs);
end;

procedure THIHiLightMemo._work_doAutoComplete;
begin
   _prop_AutoComplete:= ReadBool(_Data);
   SetOption(_prop_AutoComplete, oeAutoCompletion);
end;

procedure THIHiLightMemo._work_doOverwrite;
begin
   _prop_Overwrite:= ReadBool(_Data);
   SetOption(_prop_Overwrite, oeOverwrite);
end;

procedure THIHiLightMemo._work_doHilight;
begin
   _prop_Hilight:= ReadBool(_Data);
   SetOption(_prop_Hilight, oeHighLight);
end;

procedure THIHiLightMemo._work_doMinWordLen;
begin
   _prop_MinWordLen := ToInteger(_Data);
   AutoCompleteMinWordLength:= _prop_MinWordLen;
   Focused_CaretToView;
end;

procedure THIHiLightMemo._work_doIndent;
begin
   Indent(-_prop_Indent);
   _prop_Indent := ToInteger(_Data);
   Indent(_prop_Indent);
   Focused_CaretToView;
end;

procedure THIHiLightMemo._work_doRightMargin;
begin
   _prop_RightMargin := ReadBool(_Data);
   DrawRightMargin := _prop_RightMargin;
   Focused_CaretToView;
   Control.Invalidate;
end;

procedure THIHiLightMemo._work_doColorRightMargin;
begin
   _prop_ColorRightMargin := ToInteger(_Data);
   RightMarginColor := _prop_ColorRightMargin;
   Focused_CaretToView;
   Control.Invalidate;
end;

procedure THIHiLightMemo._var_WordAtPos;
begin
  dtString(_Data, WordAtPos(MakePoint(ToIntegerEvent(_data_WordPosition) and
           $ffff,ToIntegerEvent(_data_WordPosition) shr 16)));
end;

procedure THIHiLightMemo._var_WordAtPosMouse;
begin
  dtString(_Data, WordAtPos(FCoordinatMouse));
end;

procedure THIHiLightMemo._var_WordAtPosStartY;
begin
  dtInteger(_Data, WordAtPosStart(MakePoint(ToIntegerEvent(_data_WordPosition) and
            $ffff,ToIntegerEvent(_data_WordPosition) shr 16)).Y);
end;

procedure THIHiLightMemo._var_WordAtPosStartX;
begin
  dtInteger(_Data, WordAtPosStart(MakePoint(ToIntegerEvent(_data_WordPosition) and
            $ffff,ToIntegerEvent(_data_WordPosition) shr 16)).X);
end;

procedure THIHiLightMemo.AdjustHScroll;
var   SBInfo: TScrollInfo;
begin
   SBInfo.cbSize := Sizeof( SBInfo );
   SBInfo.fMask := SIF_PAGE or SIF_POS or SIF_RANGE;
   SBInfo.nMin := 0;
   SBInfo.nMax := Max( MaxLineWidthOnPage, Caret.X+1 );
   SBInfo.nPage := ColumnsVisiblePartial;
   SBInfo.nPos := FLeftCol;
   SBInfo.nTrackPos := FLeftCol;
   SetScrollInfo( Control.Handle, SB_HORZ, SBInfo, TRUE );
end;

procedure THIHiLightMemo.AdjustVScroll;
var SBInfo: TScrollInfo;
begin
   SBInfo.cbSize := Sizeof( SBInfo );
   SBInfo.fMask := SIF_PAGE or SIF_POS or SIF_RANGE;
   SBInfo.nMin := 0;
   SBInfo.nMax := Count-1;
   SBInfo.nPage := LinesPerPage;
   SBInfo.nPos := TopLine;
   SBInfo.nTrackPos := TopLine;
   SetScrollInfo( Control.Handle, SB_VERT, SBInfo, TRUE );
end;

procedure THIHiLightMemo.AutoAdd2Dictionary(S: PChar);
var   From: PChar;
      i, c, L, LItem: Integer;
      W: String;
      handled: Boolean;
begin
   while S^ <> #0 do begin
      if S^ in [ 'A'..'Z','a'..'z',''..'',''..'','','','_', '0'..'9' ] then begin
         From := S; inc( S );
         while S^ in [ 'A'..'Z','a'..'z',''..'',''..'','','','_', '0'..'9' ] do inc( S );
         L := DWORD( S ) - DWORD( From );
         if L > AutoCompleteMinWordLength then begin
            handled := FALSE;
            for i := 0 to FDictionary.Count-1 do begin
               c := AnsiStrLComp( From, FDictionary.ItemPtrs[ i ], L );
               if c = 0 then begin
                  if L < Length( FDictionary.Items[ i ] ) then begin
                     SetString( W, From, L );
                     FDictionary.Insert( i, W );
                  end; //       
                  handled := TRUE;
                  break;
               end
               else
                  if c > 0 then begin //    ?
                     LItem := Length( FDictionary.Items[ i ] );
                     if (LItem < L) and
                        (AnsiStrLComp( From, FDictionary.ItemPtrs[ i ], LItem ) = 0) then
                        continue; //   ,  
                  SetString( W, From, L );
                  FDictionary.Insert( i, W );
                  handled := TRUE;
                  break;
               end;
               //  c < 0,     
               //       
            end;
            if not handled then begin //    
               SetString( W, From, L );
               FDictionary.Add( W );
            end;
         end;
      end
      else inc( S );
   end;
end;

procedure THIHiLightMemo.AutoCompleteWord(n: Integer);
var   W: String;
      M: TMsg;
begin
   //  WM_CHAR,     ENTER:
   PeekMessage( M, FAutoCompletionList.Handle, WM_CHAR, WM_CHAR, pm_Remove );
   AutoCompletionHide;
   if n < 0 then Exit; //   -  
   W := FDictionary.Items[ FAutoFirst + n ];
   inc( FAutoCount );
   FDictionary.Objects[ FAutoFirst + n ] := FAutoCount;
   SelectWordUnderCursor;
   Selection := W;
   Focused_CaretToView;
end;

procedure THIHiLightMemo.AutoCompletionClear;
begin
   FDictionary.Clear;
//   Focused_CaretToView;
end;

procedure THIHiLightMemo.AutoCompletionHide;
begin
   if Assigned( FAutoCompletionForm ) and FAutoCompletionForm.Visible then begin
      FAutoCompletionForm.Hide;
      Focused_CaretToView;
   end;
end;

procedure THIHiLightMemo.AutoCompletionShow;
var   W, W1:         String;
      FromPos, Pt:   TPoint;
      DR:            TRect;
      i, j, BestIdx: Integer;
      BestCounter:   DWORD;
      DicW:          PChar;
begin
   W := WordAtPos( Caret );
   W1 := W;
   if (W = '') or not( oeAutoCompletion in Options ) then begin
      AutoCompletionHide;
      Exit;
   end;
   if Assigned( FAutoCompletionForm ) and FAutoCompletionForm.Visible and
      (FAutoWord = W) and
      (FAutoPos.X = FCaret.X) and (FAutoPos.Y = FCaret.Y) then
         Exit; //     
   FAutoPos := Caret;
   FAutoWord := W;
   AutoCompletionHide;
   FromPos := WordAtPosStart( Caret );
   W := Copy( W, 1, Caret.X - FromPos.X );
   if Length( W ) < AutoCompleteMinWordLength then Exit;

   for i := 0 to FDictionary.Count-1 do begin
      DicW := FDictionary.ItemPtrs[ i ];
      if AnsiStrLComp( PChar( W ), DicW, Length( W ) ) = 0 then begin
         FAutoFirst := i;
         FAutoCount := 0;
         BestIdx := i;
         BestCounter := 0;
         for j := i+1 to FDictionary.Count do begin
            DicW := FDictionary.ItemPtrs[ j ];
            if (j >= FDictionary.Count) or
               (AnsiStrLComp( PChar( W ), DicW, Length( W ) ) <> 0) then begin
               FAutoCount := j-i;
               break;
            end
            else
               if FDictionary.Objects[ j ] > BestCounter then begin
                  BestCounter := FDictionary.Objects[ j ];
                  BestIdx := j;
               end;
         end;
         if FAutoCount > 0 then begin
            if (FAutoCount = 1) and
               (AnsiCompareStrNoCase(FDictionary.Items[ FAutoFirst ], W1) = 0) then
                  Exit; //   1         
            if FAutoCompletionForm = nil then begin
               FAutoCompletionForm := NewForm( Applet, '' ).SetSize( 200, 200 );
               { !    Applet   
                   KOL.TApplet ! }
               FAutoCompletionForm.Visible := FALSE;
               FAutoCompletionForm.Style := WS_POPUP or WS_TABSTOP or WS_CLIPCHILDREN
                  or WS_THICKFRAME or WS_CLIPSIBLINGS;
               FAutoCompletionForm.MinHeight := 100;
               FAutoCompletionForm.OnShow := AutoFormShowEvent;
               FAutoCompletionList := NewListView( FAutoCompletionForm, lvsDetailNoHeader,
                  [ lvoRowSelect, lvoOwnerData ], nil, nil, nil ).SetAlign( caClient );
               FAutoCompletionList.LVColAdd( '', taLeft, 20 );
               FAutoCompletionList.OnMessage := AutoListMessage;
               FAutoCompletionList.OnLVData := AutoListData;
            end;
            FAutoCompletionForm.Font.Assign( Control.Font );
            FAutoCompletionList.LVCount := 0; //  -   
            Pt := MakePoint( (Caret.X - LeftCol)*CharWidth, (Caret.Y+1 - TopLine)*LineHeight );
            Pt := Control.Client2Screen( Pt );
            DR := GetDesktopRect;
            if Pt.x + FAutoCompletionForm.Width > DR.Right then
               Pt.x := Pt.x - FAutoCompletionForm.Width;
            if Pt.x < DR.Left then Pt.x := DR.Left;
            if Pt.y + FAutoCompletionForm.Height > DR.Bottom then
               Pt.y := Pt.y - LineHeight - FAutoCompletionForm.Height;
            FAutoCompletionForm.Position := Pt;
            FAutoCompletionForm.Show;
            FAutoCompletionForm.BringToFront;
            FAutoCompletionList.LVCount := FAutoCount;
            FAutoCompletionList.Focused := TRUE;
            FAutoBestFirst := BestIdx - FAutoFirst;
         end;
         Exit;
      end;
   end;
end;

{    list view ,      
     ENTER   . }

procedure THIHiLightMemo.AutoformCloseEvent(Sender: PObj; var Accept: Boolean);
begin
   Accept := FALSE;
   FAutoCompletionForm.Hide;
end;

procedure THIHiLightMemo.AutoFormShowEvent(Sender: PObj);
begin
   SendMessage(FAutoCompletionList.Handle,WM_USER+msgShow,0,0);
end;

procedure THIHiLightMemo.AutoListData(Sender: PControl; Idx, SubItem: Integer;
             var Txt: String; var ImgIdx: Integer; var State: DWORD;
var   Store: Boolean);
begin
  Txt := FDictionary.Items[ FAutoFirst + Idx ];
end;

function THIHiLightMemo.AutoListMessage(var Msg: TMsg;
            var Rslt: Integer): Boolean;
begin
   Result := FALSE;
   CASE Msg.message OF
      WM_KILLFOCUS: { ,    -    }
      begin
         FnotAdd2Dictionary1time := TRUE;
         AutoCompletionHide;
         FnotAdd2Dictionary1time := FALSE;
      end;
      WM_KEYUP, WM_SYSKEYUP, WM_SYSKEYDOWN, WM_SYSCHAR,
      WM_CHAR: {    ,   HilightMemo }
         Control.Perform( Msg.message, Msg.wParam, Msg.lParam );
      WM_KEYDOWN: {     UP, DOWN, ENTER -   
                       }
      CASE Msg.wParam OF
         VK_UP, VK_DOWN: ; //  list view 
         VK_RETURN: AutoCompleteWord( FAutoCompletionList.LVCurItem );
         VK_ESCAPE: AutoCompletionHide;
      else Control.Perform( Msg.message, Msg.wParam, Msg.lParam );
      END;
      WM_LBUTTONDOWN: // -    ,     
         SendMessage(FAutoCompletionList.Handle,WM_USER+msgComplete,0,0);
      WM_USER+msgComplete: AutoCompleteWord( FAutoCompletionList.LVCurItem );
      WM_SIZE: //   :
         FAutoCompletionList.LVColWidth[ 0 ] := FAutoCompletionList.ClientWidth - 2;
      WM_USER+msgShow: begin
         FAutoCompletionList.LVCurItem := FAutoBestFirst;
         FAutoCompletionList.LVMakeVisible( FAutoBestFirst, FALSE );
      end;
   END;
end;

function THIHiLightMemo.CanRedo: Boolean;
begin
   Result := FRedoList.Count > 0;
end;

function THIHiLightMemo.CanUndo: Boolean;
begin
   Result := FUndoList.Count > 0;
end;

procedure THIHiLightMemo.CaretToView;
begin
   if Caret.X < LeftCol then
      LeftCol := Caret.X
   else
      if LeftCol + Control.ClientWidth div CharWidth < Caret.X + 1 then
         LeftCol := Caret.X + 1 - Control.ClientWidth div CharWidth;
   if Caret.Y < TopLine then
      TopLine := Caret.Y
   else
      if TopLine + LinesPerPage <= Caret.Y then
         TopLine := Caret.Y - LinesPerPage + 1;
   Caret := Caret;
end;

procedure THIHiLightMemo.Changed;
var   y, k: Integer;
      L:    String;
begin
   if FChangeLevel <= 0 then Exit;
   Dec( FChangeLevel );
   if FChangeLevel = 0 then begin
      if not FUndoingRedoing then begin
         if FUndoList.Last[ 1 ] = 'B' then //   
         begin
           FUndoList.Delete( FUndoList.Count-1 ); Exit;
         end
         else //   
            begin
              FRedoList.Clear;
              FUndoList.Add( 'E' + Int2Str( Caret.X ) + ':' + Int2Str( Caret.Y ) );
              //      
              //    :
              k := FUndoList.Count;
              if (k >= 6) and
                 (FUndoList.Items[ k-2 ][ 1 ] = 'R') and
                 (FUndoList.Items[ k-3 ][ 1 ] = 'B') and
                 (FUndoList.Items[ k-4 ][ 1 ] = 'E') and
                 (FUndoList.Items[ k-5 ][ 1 ] = 'R') and
                 (FUndoList.Items[ k-6 ][ 1 ] = 'B') then begin
                 L := FUndoList.Items[ k-5 ];
                 Delete( L, 1, 1 );
                 y := Str2Int( L );
                 if y = Caret.Y then begin // ,     :
                    FUndoList.Delete( k-2 );
                    FUndoList.Delete( k-3 );
                    FUndoList.Delete( k-4 );
                    {       ,
                            }
                 end;
              end;
            end;
      end;
      if Assigned( Control.OnChange ) then Control.OnChange( Control );
   end;
end;

procedure THIHiLightMemo.Changing;
begin
   Inc( FChangeLevel );
   if FUndoingRedoing then Exit;
   if FChangeLevel = 1 then
      FUndoList.Add( 'B' + Int2Str( Caret.X ) + ':' + Int2Str( Caret.Y ) );
end;

function THIHiLightMemo.CharWidth: Integer;
begin
   if FCharWidth = 0 then begin
      FBitmap.Canvas.Font.Assign( Control.Font );
      FCharWidth := Control.Canvas.TextWidth( 'W' );
   end;
   Result := FCharWidth;
end;

procedure THIHiLightMemo.Clear;
var   i: Integer;
begin
  Text := '';
  FUndoList.Clear;
  FRedoList.Clear;
  for i := 0 to 9 do Bookmarks[ i ] := MakePoint( 0, 0 );
end;

function THIHiLightMemo.Client2Coord(const P: TPoint): TPoint;
begin
   Result := MakePoint( P.X div CharWidth + LeftCol,
                        P.Y div LineHeight + TopLine );
end;

function THIHiLightMemo.ColumnsVisiblePartial: Integer;
begin
   Result := Control.ClientWidth div CharWidth;
   if Result * CharWidth < Control.ClientWidth then Inc( Result );
end;

procedure THIHiLightMemo.DeleteLine(y: Integer; UseScroll: Boolean);
var   R: TRect;
begin
   if y >= Count then Exit;
   Changing;
   DeleteLine1( y );
   if (y >= TopLine) and (y < TopLine + LinesVisiblePartial) then begin
      if UseScroll and (y < TopLine + LinesPerPage) then begin
         R := MakeRect( 0, (y+1 - TopLine)*LineHeight, Control.ClientWidth,
              Control.ClientHeight );
         ScrollWindowEx( Control.Handle, 0, -LineHeight, @ R, nil, 0, nil,
            SW_INVALIDATE );
      end
      else begin
         R := MakeRect( 0, (y - TopLine)*LineHeight, Control.ClientWidth,
              Control.ClientHeight );
         InvalidateRect( Control.Handle, @ R, TRUE );
      end;
   end;
   Changed;
   FixBookmarks( MakePoint( 0, y ), 0, -1 );
   TopLine := TopLine;
end;

procedure THIHiLightMemo.DeleteLine1(y: Integer);
begin
   if y >= Count then Exit;
   if not FUndoingRedoing then
      FUndoList.Add( 'D' + Int2Str( y ) + '=' +  FLines.Items[ y ] );
   FLines.Delete( y );
end;

procedure THIHiLightMemo.DeleteSelection;
var   i, n: Integer;
      L:    String;
begin
   Changing;
   Caret := SelBegin;
   i := SelBegin.Y + 1;
   n := 0;
   while i < SelEnd.Y do begin //   
      DeleteLine1( i );
      dec( FSelEnd.Y );
      inc( n );
   end;

   L := Copy( Lines[ SelBegin.Y ], 1, SelBegin.x );
   if Length( L ) < SelBegin.x then
      L := L + StrRepeat( ' ', SelBegin.x - Length( L ) );
   L := L + CopyEnd( Lines[ SelEnd.y ], SelEnd.x + 1 );
   if SelBegin.y < SelEnd.y then begin //   
      DeleteLine( SelEnd.y, n = 0 );
   end;
   Lines[ SelBegin.y ] := L;

   if n > 0 then begin
      FixBookmarks( SelBegin, 0, -n );
      TopLine := TopLine;
      Control.Invalidate;
   end;

   SetSel( SelBegin, SelBegin, SelBegin );
   Changed;
   CaretToView;
end;

procedure THIHiLightMemo.DoKeyChar(Key: Char);
var   S, S1: String;
      t:     Integer;

   procedure InsertChar( C: Char );
   begin
      S := Lines[ Caret.Y ];
      while Length( S ) < Caret.X do S := S + ' ';
      S := Copy( S, 1, Caret.X ) + Key + CopyEnd( S, Caret.X+1 );
      Lines[ Caret.Y ] := S;
   end;

var   NeedAdd2Dictionary: Integer;
      CanShowAutoCompletion: Boolean;

begin
   if oeReadOnly in Options then Exit;
   Changing;
   NeedAdd2Dictionary := -1;
   CanShowAutoCompletion := FALSE;
TRY
   S := Lines[ Caret.Y ];
   S1 := S;
   CASE Key OF
      #27: ; // escape -  (   Autocompletion)
      #13: // enter
      begin
         if SelectionAvailable then begin
            Caret := SelBegin;
            DeleteSelection;
            if oeOverwrite in Options then begin
               CaretToView;
               Changed;
               Exit;
            end;
         end;
         if oeOverwrite in Options then begin
            //        Overwrite
            if Caret.Y = Count-1 then
               Lines[ Caret.Y+1 ] := '';
               NeedAdd2Dictionary := Caret.Y;
               Caret := MakePoint( 0, Caret.Y+1 );
         end
         else begin //    2      
            S := Lines[ Caret.Y ];
            Lines[ Caret.Y ] := Copy( S, 1, Caret.X );
            S := CopyEnd( S, Caret.X+1 );
            NeedAdd2Dictionary := Caret.Y;
            Caret := MakePoint( 0, Caret.Y+1 );
            //      SmartTab
            t := 0;
            if oeSmartTabs in Options then
               t := FindNextTabPos( Caret );
            if t > 0 then begin
               S := StrRepeat( ' ', t ) + S;
               Caret := MakePoint( t, Caret.Y );
            end;
            InsertLine( Caret.Y, S, TRUE );
         end;
         CaretToView;
      end;
      #9: // tab
      begin
         if not SelectionAvailable and (oeOverwrite in Options) then
            SetSel( Caret, MakePoint( Caret.X+1, Caret.Y ), Caret );
         if SelectionAvailable then  begin
            Caret := SelBegin;
            DeleteSelection;
            CaretToView;
         end;
         S := Lines[ Caret.Y ];
         t := Caret.X;
         if oeSmartTabs in Options then t := FindNextTabPos( Caret );
         NeedAdd2Dictionary := Caret.Y;
         if t > Caret.X then begin
            S := Copy( S, 1, Caret.X ) + StrRepeat( ' ', t - Caret.X ) +
                 CopyEnd( S, Caret.X + 1 );
            Lines[ Caret.Y ] := S;
            Caret := MakePoint( t, Caret.Y );
            SetSel( Caret, Caret, Caret );
         end
         else begin
            InsertChar( #9 );
            Caret := MakePoint( ((Caret.X + 8) div 8) * 8, Caret.Y );
         end;
      end;
      #8: // backspace
      begin
         if SelectionAvailable then begin
            Caret := SelBegin;
            DeleteSelection;
            CaretToView;
            Exit;
         end;
         if Caret.X > 0 then //  
         begin
            NeedAdd2Dictionary := Caret.Y;
            t := Caret.X;
            if (oeSmartTabs in Options) and (t > 0) and
               (Trim( Copy( S, 1, t ) ) = '') then
               t := FindPrevTabPos( Caret );
            if t < Caret.X then begin
               Delete( S, t+1, Caret.X-t );
               Lines[ Caret.Y ] := S;
               Caret := MakePoint( t, Caret.Y );
            end
            else begin
               Delete( S, Caret.X, 1 );
               Lines[ Caret.Y ] := S;
               Caret := MakePoint( Caret.X-1, Caret.Y );
            end;
         end
         else //    
         begin
            if Caret.Y = 0 then Exit; //    
               S1 := Lines[ Caret.Y-1 ];
            Caret := MakePoint( Length( S1 ), Caret.Y-1 );
            NeedAdd2Dictionary := Caret.Y;
            S1 := S1 + S;
            Lines[ Caret.Y ] := S1;
            DeleteLine( Caret.Y+1, TRUE );
         end;
      end;
      else // any char (edit)
         if Key < ' ' then Exit; //    
      if not SelectionAvailable and (oeOverwrite in Options) then
         SetSel( Caret, MakePoint( Caret.X+1, Caret.Y ), Caret );
      if SelectionAvailable then begin
         Caret := SelBegin;
         DeleteSelection;
         CaretToView;
      end;
      if Key in [ 'A'..'Z','a'..'z',''..'',''..'','','','_', '0'..'9' ] then
         CanShowAutocompletion := TRUE
      else
         NeedAdd2Dictionary := Caret.Y;
      InsertChar( Key );
      Caret := MakePoint( Caret.X+1, Caret.Y );
   END;
   SetSel( Caret, Caret, Caret );
FINALLY
   Changed;
   CaretToView;
END;
   if (NeedAdd2Dictionary >= 0) and (NeedAdd2Dictionary < Count) then
      AutoAdd2Dictionary( FLines.ItemPtrs[ NeedAdd2Dictionary ] );
   if CanShowAutocompletion then
      AutoCompletionShow
   else
      AutoCompletionHide;
end;

procedure THIHiLightMemo.DoKeyDown(Key, Shift: Integer);
   procedure MoveLeftUp( NewCaretX, NewCaretY: Integer );
   begin
      if (NewCaretX < 0) and (NewCaretY > 0) then begin
         Dec( NewCaretY );
         NewCaretX := Length( Lines[ NewCaretY ] );
      end;
      Caret := MakePoint( Max( 0, NewCaretX ),
                          Min( Count-1, Max( 0, NewCaretY ) ) );
      CaretToView;
      if Shift and MK_SHIFT <> 0 then SetSel( Caret, SelFrom, SelFrom )
      else SetSel( Caret, Caret, Caret );
   end;

   procedure MoveRightDown( NewCaretX, NewCaretY: Integer );
   begin
      Caret := MakePoint( Max( 0, NewCaretX ),
                          Min( Count-1, Max( 0, NewCaretY ) ) );
      CaretToView;
      if Shift and MK_SHIFT <> 0 then SetSel( SelFrom, Caret, SelFrom )
      else SetSel( Caret, Caret, Caret );
   end;
   
var   NewCaret: TPoint;
      S:        String;
begin
   Changing;
TRY
   CASE Key OF
   VK_LEFT:      if Shift and MK_CONTROL <> 0 then begin
                    NewCaret := WordAtPosStart( Caret );
                    NewCaret := FindPrevWord( NewCaret, TRUE );
                    MoveLeftUp   ( NewCaret.X, NewCaret.Y );
                 end
                 else
                    MoveLeftUp ( Caret.X-1, Caret.Y );
   VK_RIGHT:
                 if Shift and MK_CONTROL <> 0 then begin
                    NewCaret := WordAtPosStart( Caret );
                    Inc( NewCaret.X, Length( WordAtPos( Caret ) ) );
                    NewCaret := FindNextWord( NewCaret, TRUE );
                    MoveRightDown( NewCaret.X, NewCaret.Y );
                 end
                 else
                    MoveRightDown( Caret.X+1, Caret.Y );
   VK_HOME:      if Shift and MK_CONTROL <> 0 then MoveLeftUp ( 0, 0 )
                 else MoveLeftUp ( 0, Caret.Y );
   VK_END:       if Shift and MK_CONTROL <> 0 then MoveRightDown( 0, Count-1 )
                 else MoveRightDown( Length( TrimRight( Lines[ Caret.Y ] ) ),
                                     Caret.Y );
   VK_UP:        MoveLeftUp   ( Caret.X, Caret.Y-1 );
   VK_DOWN:      MoveRightDown( Caret.X, Caret.Y+1 );
   VK_PAGE_UP:   MoveLeftUp   ( Caret.X, Caret.Y - LinesPerPage );
   VK_PAGE_DOWN: MoveRightDown( Caret.X, Caret.Y + LinesPerPage );
   VK_DELETE:    if oeReadOnly in Options then begin
                    Changed; Exit;
                 end else if SelectionAvailable then begin
                    if Shift and MK_SHIFT <> 0 then
                       Control.Perform( WM_CUT, 0, 0 )
                    else begin
                       Caret := SelBegin;
                       DeleteSelection;
                    end;
                 end else if Caret.X >= Length( Lines[ Caret.Y ] ) then begin //  
                    if Caret.Y = Count-1 then Exit;
                    S := Lines[ Caret.Y ];
                    while Length( S ) < Caret.X do S := S + ' ';
                    Lines[ Caret.Y ] := S + Lines[ Caret.Y+1 ];
                    DeleteLine( Caret.Y+1, TRUE );
                    SetSel( Caret, Caret, Caret );
                 end else begin //   
                    S := Lines[ Caret.Y ];
                    Delete( S, Caret.X+1, 1 );
                    Lines[ Caret.Y ] := S;
                 end;
    VK_INSERT:   if oeReadOnly in Options then Exit
                 else if Shift and MK_SHIFT <> 0 then
                    Control.Perform( WM_PASTE, 0, 0 )
                 else if Shift = 0 then begin
                     if oeOverwrite in Options then
                        Options := Options - [ oeOverwrite ]
                     else
                        Options := Options + [ oeOverwrite ];
                     Caret := Caret;
                 end;
    Word( 'V' ): if oeReadOnly in Options then Exit
                 else if Shift and MK_CONTROL <> 0 then
                    Control.Perform( WM_PASTE, 0, 0 );
    Word( 'Y' ): if oeReadOnly in Options then Exit
                 else if Shift and MK_CONTROL <> 0 then begin
                    DeleteLine( Caret.Y, TRUE );
                    Caret := Caret;
                 end;
    Word( 'X' ): if oeReadOnly in Options then Exit
                 else if Shift and MK_CONTROL <> 0 then
                    Control.Perform( WM_CUT, 0, 0 );
    Word( 'C' ): if Shift and MK_CONTROL <> 0 then
                    Control.Perform( WM_COPY, 0, 0 );
    Word( 'A' ): if Shift and MK_CONTROL <> 0 then
                    SetSel(MakePoint(0, 0), MakePoint(0,FLines.Count), MakePoint(0, 0));
    Word( 'Z' ): begin
                    if oeReadOnly in Options then Exit;
                    while FChangeLevel > 0 do Changed;
                    {   B<x>:<y>,   Changing }
                    if Shift and MK_CONTROL <> 0 then begin
                       if Shift and MK_SHIFT <> 0 then // ctrl+shift+Z - redo
                          Redo
                       else
                          Undo;
                    end;
                    Exit; {   Changed }
                 end;
    END;
FINALLY
   Changed;
END;
   if Key in [Word('0')..Word('9')] then
      if Shift and MK_CONTROL <> 0 then begin
         if not Shift and MK_SHIFT <> 0 then // ctrl+shift+(1..9)
            Bookmarks[Ord(Key) - Ord('0')] := Caret
         else begin
            if (Bookmarks[Ord(Key) - Ord('0')].X and Bookmarks[Ord(Key) - Ord('0')].Y) = 0 then exit;  
            Caret:= Bookmarks[Ord(Key) - Ord('0')];
            CaretToView;
         end;
      end;
end;

procedure THIHiLightMemo.DoMouseDown(X, Y: Integer; Shift: Integer);
var   Pt:    TPoint;
      Attrs: TTokenAttrs;
      S:     string;
      i, j:  integer;
begin
   Pt := MakePoint(
         X div CharWidth + LeftCol,
         Y div LineHeight + TopLine );
   if Shift and MK_SHIFT <> 0 then SetSel( SelFrom, Pt, SelFrom )
   else SetSel( Pt, Pt, Pt );
   Caret := Pt;
   FMouseDown := TRUE;
   SetCapture( Control.Handle );
   if not ((oeReadOnly in Options) and (oeHighLight in Options)) then Exit;
   S := Lines[SelFrom.Y];
   if S = '' then exit;
   i := 0;
   while (i < Length(S)) do begin
      onScanToken(Control, MakePoint(i, SelFrom.Y), Attrs);     
      if Attrs.fontstyle = TFontStyle({$ifdef F_P}integer(_prop_HilightFont.Style)
                                      {$else}     _prop_HilightFont.Style
                                      {$endif})   + [fsUnderline] then begin 
         j := i;
         if (S[ j+1 ] <= ' ') then
            while (j < Length(S)) and (S[ j+1 ] <= ' ') do inc(j)
         else
            while (j < Length(S))  and (S[ j+1 ] > ' ') do inc(j);
         if (SelFrom.X >= i) and (SelFrom.X < j) then begin
            S := Copy( S, i, j - i + 1);        
            while (S[Length(S)] in ['!'..'/',':'..'?','['..'`']) do deleteTail(S,1);
            _hi_OnEvent(_event_onClickUnderLineStr, S);
            exit;
         end;
         i := j + 1;
         Continue;
      end
      else
         inc(i); 
   end;
end;

procedure THIHiLightMemo.DoMouseMove(X, Y: Integer);
var   Pt: TPoint;
begin
   Pt := MakePoint(
         X div CharWidth + LeftCol,
         Y div LineHeight + TopLine );
   FCoordinatMouse := Pt;
   if not FMouseDown then begin
      if not ((oeReadOnly in Options) and (oeHighLight in Options)) then Exit;
      Control.Invalidate;
      Exit; // todo:      ?
   end;
   SetSel( SelFrom, Pt, SelFrom );
   Caret := Pt;
   CaretToView;
end;

procedure THIHiLightMemo.DoPaint(DC: HDC);
var   x, y, i, L:        Integer;
      R, R0, Rsel, CR:   TRect;
      OldClip, NewClip:  HRgn;
      P, MPSTART, MPEND: TPoint;
      Attrs:             TTokenAttrs;
      Canvas:            PCanvas;
begin
   CR := Control.ClientRect;
   NewClip := CreateRectRgnIndirect( Control.ClientRect );
   CombineRgn( NewClip, NewClip, Control.UpdateRgn, RGN_AND );
   SelectClipRgn( DC, NewClip );
   DeleteObject( NewClip );
   y := 0;
   if (FBufferBitmap <> nil) and (
      (FBufferBitmap.Width < Control.ClientWidth) or
      (FBufferBitmap.Height < LineHeight) ) then
      Free_And_Nil( FBufferBitmap );
   if FBufferBitmap = nil then
      FBufferBitmap := NewBitmap( Control.ClientWidth, LineHeight );
   Canvas := FBufferBitmap.Canvas;
   Canvas.Font.Assign( Control.Font );
   for i := TopLine to TopLine + LinesPerPage + 1 do begin
      if y >= CR.Bottom then break;
      if i >= Count then break;
      R := MakeRect( 0, y, CR.Right, y + LineHeight );
      R0 := R;
      OffsetRect( R0, -R0.Left, -R0.Top );
      if SelectionAvailable and (FSelBegin.Y <= i) and (FSelEnd.Y >= i) then
      begin //        :
         Control.Canvas.Brush.Color := clHighlight;
         Control.Canvas.Font.Color := clHighlightText;
         Control.Canvas.Font.FontStyle := [ ];
         Rsel := R;
         if i = SelBegin.Y then
            Rsel.Left := Max( 0, (FSelBegin.X - FLeftCol)*CharWidth );
         if i = SelEnd.Y then
            Rsel.Right := Max( 0, (FSelEnd.X - FLeftCol)*CharWidth );
         if Rsel.Right > Rsel.Left then begin
            OldClip := CreateRectRgn( 0,0,0,0 );
            GetClipRgn( DC, OldClip );
            NewClip := CreateRectRgnIndirect( Rsel );
            SelectClipRgn( DC, NewClip );
            Control.Canvas.FillRect(R);
            P := MakePoint( 0, i );
            while P.X < LeftCol + ColumnsVisiblePartial do begin
               Control.Canvas.TextOut( R0.Left + (P.X - LeftCol)*CharWidth, R.Top,
                                       Copy( Lines[ i ], P.X+1, 1 ) );
               Inc( P.X );
            end;
            SelectClipRgn( DC, OldClip );
            ExtSelectClipRgn( DC, NewClip, RGN_DIFF );
            DeleteObject( NewClip );
            DeleteObject( OldClip );
         end;
      end;
      if RectInRegion( Control.UpdateRgn, R ) then begin
         Canvas.Brush.Color := Control.Color;
         Canvas.Font.Color := _prop_Font.Color;
         SetFont(Canvas.Font,_prop_Font.Style);
         if Assigned( FOnScanToken ) and (oeHighlight in Options) then begin
            Canvas.FillRect( R0 );
            Canvas.Brush.BrushStyle := bsClear;
            P := MakePoint( 0, i );
            Attrs.fontcolor := _prop_Font.Color;
            Attrs.fontstyle := TFontStyle({$ifdef F_P}integer(_prop_Font.Style)
                                          {$else}_prop_Font.Style{$endif});
            while P.X < LeftCol+ColumnsVisiblePartial do begin
               L := OnScanToken( Control, P, Attrs );
               MPSTART := P;
               MPEND := P;
               MPEND.X := P.X + L;
               if L <= 0 then L := Length( Lines[ i ] ) - P.X;
               if L <= 0 then break;
               if P.X + L >= LeftCol then begin
                  Canvas.Font.FontStyle := Attrs.fontstyle;
                  while L > 0 do begin
                     if (FCoordinatMouse.Y = MPSTART.Y) and (FCoordinatMouse.X >= MPSTART.X) and
                        (FCoordinatMouse.X < MPEND.X) and (Attrs.fontstyle = TFontStyle(
                        {$ifdef F_P}integer(_prop_HilightFont.Style)
                        {$else}     _prop_HilightFont.Style
                        {$endif})   + [fsUnderline]) then
                        Canvas.Font.Color := _prop_ColorUnderLine
                     else   
                        Canvas.Font.Color := Attrs.fontcolor;
                     Canvas.TextOut( R0.Left + (P.X - LeftCol)*CharWidth, R0.Top,
                                     Copy( Lines[ i ], P.X+1, 1 ) );
                     Inc( P.X );
                     Dec( L );
                  end;
               end
               else P.X:= P.X + L;
            end;
            Canvas.Brush.BrushStyle := bsSolid;
            if FRightMargin then begin
               x := (RightMarginChars - LeftCol)*CharWidth;
               if Length( TrimRight( Lines[ i ] ) ) > RightMarginChars then begin
                  Canvas.Pen.Color := RightMarginColor;
                  Canvas.MoveTo( x, R0.Top );
                  Canvas.LineTo( x-2, R0.Top + LineHeight div 3 );
                  Canvas.LineTo( x+2, R0.Top + LineHeight * 2 div 3 );
                  Canvas.LineTo( x, R0.Bottom );
               end
               else begin
                  Canvas.Brush.Color := RightMarginColor;
                  Canvas.FillRect( MakeRect( x, R0.Top, x+1, R0.Bottom ) );
                  Canvas.Brush.Color := Control.Color;
               end;
               end;
            BitBlt( DC, R.Left, R.Top, R.Right - R.Left, R.Bottom - R.Top,
                    Canvas.Handle, 0, 0, SRCCOPY );
         end
         else begin
            Control.Canvas.Brush.Color := Control.Color;
            Canvas.Font.Color := _prop_Font.Color;
            SetFont(Canvas.Font,_prop_Font.Style);
            Control.Canvas.TextRect( R, 0, R.Top, CopyEnd( Lines[ i ], LeftCol + 1 ) );
            if FRightMargin then begin
               x := (RightMarginChars - LeftCol)*CharWidth;
               if Length( TrimRight( Lines[ i ] ) ) > RightMarginChars then begin
                  Control.Canvas.Pen.Color := RightMarginColor;
                  Control.Canvas.MoveTo( x, R.Top );
                  Control.Canvas.LineTo( x-2, R.Top + LineHeight div 3 );
                  Control.Canvas.LineTo( x+2, R.Top + LineHeight * 2 div 3 );
                  Control.Canvas.LineTo( x, R.Bottom );
               end
               else Control.Canvas.Brush.Color := RightMarginColor;
               Control.Canvas.FillRect( MakeRect( x, R.Top, x+1, R.Bottom ) );
               Control.Canvas.Brush.Color := Control.Color;
            end;
         end;
         //GdiFlush;
      end;
      y := y + LineHeight;
   end;
   if y < CR.Bottom then begin //     
      Control.Canvas.Brush.Color := Control.Color;
      if FRightMargin then begin
         x := (RightMarginChars - LeftCol)*CharWidth;
         if x < CR.Right then begin
            R := MakeRect( 0, y, x, CR.Bottom );
            Control.Canvas.FillRect( R );
            Control.Canvas.Brush.Color := RightMarginColor;
            R.Left := x;
            R.Right := x+1;
            Control.Canvas.FillRect( R );
            Control.Canvas.Brush.Color := Control.Color;
            R.Left := x+1;
            R.Right := CR.Right;
            Control.Canvas.FillRect( R );
         end
         else 
            Control.Canvas.FillRect( MakeRect( 0, y, CR.Right, CR.Bottom ) );
      end
      else 
         Control.Canvas.FillRect( MakeRect( 0, y, CR.Right, CR.Bottom ) );
   end;
end;

procedure THIHiLightMemo.DoScroll(Cmd, wParam: DWord);
begin
   CASE Cmd OF
      SC_HSCROLL:
         CASE loWord( wParam ) OF
            SB_LEFT, SB_LINELEFT:  LeftCol := LeftCol-1;
            SB_RIGHT,SB_LINERIGHT: LeftCol := LeftCol+1;
            SB_PAGELEFT:           LeftCol := LeftCol - Control.ClientWidth div CharWidth;
            SB_PAGERIGHT:          LeftCol := LeftCol + Control.ClientWidth div CharWidth;
            SB_THUMBTRACK:         LeftCol := HiWord( wParam );
         END;
     SC_VSCROLL:
        CASE loWord( wParam ) OF
           SB_LEFT, SB_LINELEFT:  TopLine := TopLine-1;
           SB_RIGHT,SB_LINERIGHT: TopLine := TopLine+1;
           SB_PAGELEFT:           TopLine := TopLine - LinesPerPage;
           SB_PAGERIGHT:          TopLine := TopLine + LinesPerPage;
           SB_THUMBTRACK:         TopLine := HiWord( wParam );
        END;
   END;
end;

procedure THIHiLightMemo.DoUndoRedo(List1, List2: PStrList);
var   L1, L2: String;
      x, y:   Integer;
begin
   // :     List1,
   //      List2
   FUndoingRedoing := TRUE;
   Assert( List1.Last[ 1 ] = 'E' );
   List2.Add( 'B' + CopyEnd( List1.Last, 2 ) );
   List1.Delete( List1.Count-1 );
   while TRUE do begin
      L1 := List1.Last;
      List1.Delete( List1.Count-1 );
      CASE L1[ 1 ] OF
      'A': // :    , :   
      begin
         L2 := 'D' + Int2Str( Count-1 ) + '=' + FLines.Last;
         FLines.Delete( Count-1 );
      end;
     'D': // :  , :  
      begin
         Delete( L1, 1, 1 );
         y := Str2Int( L1 );
         Parse( L1, '=' );
         L2 := 'I' + Int2Str( y );
         InsertLine( y, L1, FALSE );
      end;
     'I': // :  , :  
      begin
         Delete( L1, 1, 1 );
         y := Str2Int( L1 );
         L2 := 'D' + Int2Str( y ) + '=' + Lines[ y ];
         DeleteLine( y, FALSE );
      end;
     'R': // :  
      begin
         Delete( L1, 1, 1 );
         y := Str2Int( L1 );
         Parse( L1, '=' );
         L2 := 'R' + Int2Str( y ) + '=' + Lines[ y ];
         Lines[ y ] := L1;
      end;
     'B': //     
      begin
         L2 := 'E' + CopyEnd( L1, 2 );
         Delete( L1, 1, 1 );
         x := Str2Int( L1 );
         Parse( L1, ':' );
         y := Str2Int( L1 );
         //        
         Caret := MakePoint( x, y );
         SetSel( Caret, Caret, Caret );
         CaretToView;
      end;
      END;
      List2.Add( L2 );
      if L2[ 1 ] = 'E' then break;
   end;
   FUndoingRedoing := FALSE;
   AdjustHScroll;
   AdjustVScroll;
   //  ,  OnChange:
   if Assigned( Control.OnChange ) then
      Control.OnChange( Control );
end;

procedure THIHiLightMemo.FastFillDictionary;
type
   TByteArray = array[ 0..100000 ] of Byte;
   PByteArray = ^TByteArray;
var   i:             Integer;
      HashTable:     PByteArray;
      S, From:       PChar;
      TempMemStream: PStream;
      ChkSum:        DWORD;
      EOL:           Char;
      DicAsTxt:      String;
begin
   FDictionary.Clear;
   HashTable := AllocMem( 65536 ); //   512 1-  ""
   TempMemStream := NewMemoryStream; //    
   EOL := #13;
TRY
   for i := 0 to FLines.Count-1 do begin
      S := FLines.ItemPtrs[ i ];
      while S^ <> #0 do  begin
         if S^ in [ 'A'..'Z','a'..'z',''..'',''..'','','','_', '0'..'9' ] then begin
            From := S;
            ChkSum := Byte( From^ );
            inc( S );
            while S^ in [ 'A'..'Z','a'..'z',''..'',''..'','','','_', '0'..'9' ] do begin
                ChkSum := (( ChkSum shl 1) or (ChkSum shr 18) and 1 ) xor Byte( Upper[ S^ ] );
                inc( S );
            end;
            if DWORD(S) - DWORD(From) > DWORD(AutoCompleteMinWordLength) then begin
               ChkSum := ChkSum and $7FFFF;
               if HashTable[ ChkSum shr 3 ] and (1 shl (ChkSum and 7)) = 0 then begin
                  HashTable[ ChkSum shr 3 ] := HashTable[ ChkSum shr 3 ] or (1 shl (ChkSum and 7));
                  TempMemStream.Write( From^, DWORD( S ) - DWORD( From ) );
                  TempMemStream.Write( EOL, 1 );
               end;
            end;
         end
         else Inc( S );
      end;
   end;
   if TempMemStream.Position > 0 then begin
      SetString( DicAsTxt, PChar( TempMemStream.Memory ),
                 TempMemStream.Position );
      FDictionary.Text := DicAsTxt;
      FDictionary.Sort( FALSE );
   end;
FINALLY
   TempMemStream.Free;
   FreeMem( HashTable );
END;
end;

function THIHiLightMemo.FindNextTabPos(const FromPos: TPoint): Integer;
var   P, P1, P2: TPoint;
begin
   P := FromPos;
   P1 := P;
   Result := P.X;
   while (P.Y > 0) or (P.X > 0) do begin
      P2 := P;
      P := FindPrevWord( P2, FALSE );
      if (P.X = P2.X) and (P.Y = P2.Y) then begin
         Result := P.X;
         break;
      end
      else if (P.X >= FromPos.X) and ((Result = FromPos.X) or (P.Y = P1.Y)) then begin
         Result := P.X;
         P1 := P;
      end
      else if (P.Y < P1.Y) or (P.X <= FromPos.X) then  break;
   end;
end;

function THIHiLightMemo.FindNextWord(const FromPos: TPoint;
         LookForLettersDigits: Boolean): TPoint;
var   S:    String;
      i, y: Integer;
begin
   y := FromPos.Y;
   i := FromPos.X;
   while TRUE do begin
      S := Lines[ y ];
      while i < Length( S )-1 do begin
         inc( i );
         if LookForLettersDigits and IsLetterDigit( S[ i+1 ] ) or
            not LookForLettersDigits and (S[ i+1 ] > ' ') then begin //   
            Result := MakePoint( i, y );
            Exit;
         end;
      end;
      //    ,    
      if y = Count-1 then //   
         break;
      i := -1;
      inc( y );
   end;
   Result := MakePoint( i, y );
end;

function THIHiLightMemo.FindPrevTabPos(const FromPos: TPoint): Integer;
var   P, P2: TPoint;
begin
   P := FromPos;
   Result := P.X;
   while (P.Y > 0) or (P.X > 0) do begin
      P2 := P;
      P := FindPrevWord( P, FALSE );
      if (P.X = P2.X) and (P.Y = P2.Y) or (P.X < FromPos.X) then begin
         Result := P.X;
         break;
      end;
   end;
end;

function THIHiLightMemo.FindPrevWord(const FromPos: TPoint;
         LookForLettersDigits: Boolean): TPoint;
var   S:    String;
      i, y: Integer;
begin
   Result := FromPos;
   y := FromPos.Y;
   i := FromPos.X;
   while TRUE do begin
      S := Lines[ y ];
      if i = -1 then i := Length( S )+1
      else dec( i );
      while i > 0 do begin
         dec( i );
         if (LookForLettersDigits and
             not IsLetterDigit( (Copy( S, i+1, 1 ) + ' ')[ 1 ] ) or
             not LookForLettersDigits and (Copy( S, i+1, 1 ) <= ' '))
             and (i+2 <= Length( S ) ) and (LookForLettersDigits and
             IsLetterDigit( (Copy( S, i+2, 1 ) + ' ')[ 1 ] ) or
             not LookForLettersDigits and (Copy( S, i+2, 1 ) > ' ')) then begin //   
            Result := MakePoint( i+1, y );
            Exit;
         end
         else if (i = 0) and (Length( S ) > 0) and IsLetterDigit( S[ 1 ] ) then begin
            Result := MakePoint( 0, y );
            Exit;
         end;
      end;
      //    ,    
      if y = 0 then //   
         Exit;
      i := -1;
      dec( y );
   end;
   Result := MakePoint( i, y );
end;

function THIHiLightMemo.FindReplace( S: String; const replaceTo: string;
         const FromPos: TPoint; FindReplaceOptions: TFindReplaceOptions;
         SelectFound: Boolean): TPoint;
var   P: TPoint;
      L: String;
      i: Integer;

   function CompareSubstr( var i: Integer ): Boolean;
   var   j: Integer;
   begin
      Result := FALSE;
      i := 1;
      j := 1;
      while j <= Length( S ) do
      begin
         if P.X + i > Length( L ) then Exit;
         if not(froSpaces in FindReplaceOptions) and (S[ j ] = ' ') then begin
            inc( j );
            if not(L[ P.X + i ] in [ #9, ' ' ]) then Exit;
            while (P.X + i <= Length( L )) and
                  (L[ P.X + i ] in [ #9, ' ' ]) do inc( i );
         end
         else if froCase in FindReplaceOptions then begin
            if L[ P.X + i ] <> S[ j ] then Exit;
            inc( i );
            inc( j );
         end
         else begin
            if Upper[ L[ P.X + i ] ] <> Upper[ S[ j ] ] then Exit;
            inc( i );
            inc( j );
         end;
      end;
      Result := TRUE;
   end;

begin
   Result := FromPos;
   if not( froSpaces in FindReplaceOptions ) then begin
      while StrReplace( S, #9, ' ' ) do;
      while StrReplace( S, '  ', ' ' ) do;
   end;
   P := FromPos;
   L := Lines[ P.Y ];
   while TRUE do begin
      if froBack in FindReplaceOptions then begin
        Dec( P.X );
        if P.X < 0 then begin
           Dec( P.Y );
           L := Lines[ P.Y ];
           if P.Y < 0 then Exit;
           P.X := Length( L );
           continue;
        end;
      end
      else begin
         Inc( P.X );
         if P.X >= Length( L ) then begin
            P.X := 0;
            Inc( P.Y );
            if P.Y >= Count then Exit;
            L := Lines[ P.Y ];
         end;
      end;
      if CompareSubstr( i ) then begin
         if froReplace in FindReplaceOptions then begin
            SelBegin := P;
            SelEnd := MakePoint( P.X+i-1, P.Y );
            Selection := ReplaceTo;
            SelEnd := Caret;
         end
         else if SelectFound then begin
            SelBegin := P;
            SelEnd := MakePoint( P.X+i-1, P.Y );
            Caret := SelEnd;
         end;
         if not( froReplace in FindReplaceOptions ) or
            not( froAll in FindReplaceOptions ) then
         break;
      end;
   end;
   CaretToView;
   Result := P;
end;

procedure THIHiLightMemo.FixBookmarks(const FromPos: TPoint; deltaX, deltaY: Integer);
var   i: Integer;
begin
   for i := 0 to 9 do
      if (FBookmarks[ i ].X <> 0) or (FBookmarks[ i ].Y <> 0) then begin
         if (FBookmarks[ i ].Y = FromPos.Y) and (FBookmarks[ i ].X >= FromPos.X) then
            Inc( FBookmarks[ i ].X, deltaX );
         if (FBookmarks[ i ].Y > FromPos.Y) then
            Inc( FBookmarks[ i ].Y, deltaY );
      end;
end;

function THIHiLightMemo.GetBookMark(IDx: Integer): TPoint;
begin
   Result := FBookmarks[ Idx ];
end;

function THIHiLightMemo.GetCount: Integer;
begin
   Result := FLines.Count;
end;

function THIHiLightMemo.GetLines(Idx: Integer): String;
var   i, j: Integer;
begin
   Result := FLines.Items[ Idx ];
   //        :
   i := 0;
  while i < Length( Result ) do begin
     if Result[ i+1 ] = #9 then begin
        j := ((i + 8) div 8) * 8;
        Result := Copy( Result, 1, i ) + StrRepeat( ' ', j-i ) +
                  CopyEnd( Result, i+2 );
     end;
     inc( i );
  end;
  if not( oeKeepTrailingSpaces in Options) then
     Result := TrimRight( Result );
end;

function THIHiLightMemo.GetSelection: String;
var   i: Integer;
      S: String;
begin
   Result := '';
   //        
   for i := SelBegin.Y to SelEnd.Y do begin
      S := Lines[ i ];
      if i = SelEnd.Y then
         S := Copy( S, 1, SelEnd.X );
      if i = SelBegin.Y then
         S := CopyEnd( S, SelBegin.X+1 );
      if i > SelBegin.Y then
         Result := Result + #13#10;
      Result := Result + S;
   end;
end;

function THIHiLightMemo.GetText: String;
begin
   Result := FLines.Text;
end;

function THIHiLightMemo.GetDrawRightMargin: Boolean;
begin
   Result := FRightMargin;
end;

procedure THIHiLightMemo.Indent(delta: Integer);
var   y, i, k: Integer;
      S:       String;
begin
   Changing;
   for y := FSelBegin.y to FSelEnd.y do begin
      if (FSelEnd.y > FSelBegin.y) and (FSelEnd.y = y) and (FSelEnd.x = 0) then
         break;
      S := Lines[ y ];
      if delta > 0 then begin
         S := StrRepeat( ' ', delta ) + S;
         FixBookmarks( MakePoint( 0, y ), delta, 0 );
      end
      else if delta < 0 then begin
         k := 0;
         for i := 1 to Length( S ) do
            if S[ i ] > ' ' then break
            else inc( k );
         if -delta < k then k := -delta;
         if k > 0 then begin
            Delete( S, 1, k );
            FixBookmarks( MakePoint( 0, y ), -k, 0 );
         end;
      end
      else
         break;
      Lines[ y ] := S;
   end;
   if (FSelEnd.x = 0) and (FSelEnd.y > FSelBegin.y) then
   SetSel( MakePoint( FSelBegin.x + delta, FSelBegin.y ),
           FSelEnd,
           MakePoint( FSelBegin.x + delta, FSelBegin.y ) )
   else
   SetSel( MakePoint( FSelBegin.x + delta, FSelBegin.y ),
           MakePoint( FSelEnd.x + delta, FSelEnd.y ),
           MakePoint( FSelFrom.x + delta, FSelBegin.y ) );
   Changed;
   CaretToView;
end;

procedure THIHiLightMemo.InsertLine(y: Integer; const S: String; UseScroll: Boolean);
var   R: TRect;
begin
   Changing;
   while Count < y do begin
      FLines.Add( '' );
      if not FUndoingRedoing then
         FUndoList.Add( 'A' );
      UseScroll := FALSE;
   end;
   FLines.Insert( y, S );
   FixBookmarks( MakePoint( 0, y ), 0, 1 );
   if not FUndoingRedoing then
      FUndoList.Add( 'I' + Int2Str( y ) );
   if UseScroll and (y >= TopLine) and (y < TopLine + LinesPerPage) then begin
      R := MakeRect( 0, (y - TopLine)*LineHeight, Control.ClientWidth,
           Control.ClientHeight );
      ScrollWindowEx( Control.Handle, 0, LineHeight, @ R, nil, 0, nil,
           SW_INVALIDATE );
   end
   else begin
      R := MakeRect( 0, (y - TopLine)*LineHeight, Control.ClientWidth,
           Control.ClientHeight );
      InvalidateRect( Control.Handle, @ R, TRUE );
   end;
   Changed;
end;

procedure THIHiLightMemo.InvalidateLine(y: Integer);
var   R: TRect;
begin
   if y < TopLine then Exit;
   if y > TopLine + LinesVisiblePartial then Exit;
   R := MakeRect( 0, (y - TopLine)*LineHeight, Control.ClientWidth,
                     (y+1 - TopLine)*LineHeight );
   InvalidateRect( Control.Handle, @ R, TRUE );
end;

procedure THIHiLightMemo.InvalidateSelection;
var   y: Integer;
      R: TRect;
begin
   //    ,  ,  ,  
   //  ,    -  .
   for y := Max( TopLine, SelBegin.Y ) to
            Min( TopLine + LinesVisiblePartial - 1, SelEnd.Y ) do begin
       R := MakeRect( 0, (y - TopLine)*LineHeight, Control.ClientWidth,
                      Control.ClientHeight );
       InvalidateRect( Control.Handle, @ R, TRUE );
   end;
end;

function THIHiLightMemo.LineHeight: Integer;
begin
   if FLineHeight = 0 then begin
      FBitmap.Canvas.Font.Assign( Control.Font );
      FLineHeight := FBitmap.Canvas.TextHeight( 'A/_' );
   end;
   Result := FLineHeight;
   if Result = 0 then Result := 16;
end;

function THIHiLightMemo.LinesPerPage: Integer;
begin
   Result := Control.ClientHeight div LineHeight;
end;

function THIHiLightMemo.LinesVisiblePartial: Integer;
begin
   Result := Control.ClientHeight div LineHeight;
   if Result * LineHeight < Control.ClientHeight then
      Inc( Result );
end;

function THIHiLightMemo.MaxLineWidthOnPage: Integer;
var   i: Integer;
begin
   Result := 0;
   for i := TopLine to TopLine + (Control.Height+LineHeight-1) div LineHeight - 1 do begin
      if i >= Count then break;
      Result := Max( Result, Length( Lines[ i ] ) );
  end;
end;

procedure THIHiLightMemo.Redo;
begin
   if not CanRedo then Exit;
   DoUndoRedo( FRedoList, FUndoList );
end;

function THIHiLightMemo.SelectionAvailable: Boolean;
begin
   Result := ((SelBegin.X <> SelEnd.X) or (SelBegin.Y <> SelEnd.Y)) and
             ((SelBegin.Y < Count-1) or
             (SelBegin.Y = Count-1) and
             (SelBegin.X < Length( FLines.Last )));
end;

procedure THIHiLightMemo.SelectWordUnderCursor;
var   WordStart, WordEnd: TPoint;
      W:                  String;
begin
   W := WordAtPos( Caret );
   if W = '' then Exit;
   WordStart := WordAtPosStart( Caret );
   WordEnd := WordStart;
   inc( WordEnd.X, Length( W ) );
   SetSel( WordStart, WordEnd, WordStart );
   Caret := WordEnd;
   CaretToView;
end;

procedure THIHiLightMemo.SetBookmark( Idx: Integer; const Value: TPoint );
begin
   FBookmarks[ Idx ] := Value;
end;

procedure THIHiLightMemo.SetCaret(const Value: TPoint);
begin
   if (FCaret.Y < Count) and (FCaret.Y >= 0) and (FCaret.Y <> Value.Y) and
       not FnotAdd2Dictionary1time then
     AutoAdd2Dictionary( FLines.ItemPtrs[ FCaret.Y ] );
   if (FCaret.x <> Value.x) or (FCaret.y <> Value.y) then
      AutoCompletionHide;
   FCaret := Value;
   if FCaret.Y < 0 then FCaret.Y := 0; 
   if FCaret.X < 0 then FCaret.X := 0;      
   if (FCaret.Y >= TopLine) and
      (FCaret.Y < TopLine + LinesVisiblePartial) and
      (FCaret.X >= LeftCol) and
      (FCaret.X <= LeftCol + ColumnsVisiblePartial) and
      (GetFocus = Control.Handle) then begin
      if oeOverwrite in Options then
         CreateCaret( Control.Handle, 0, CharWidth, LineHeight )
      else
         CreateCaret( Control.Handle, 0, 1, LineHeight );
      SetCaretPos( (FCaret.X - LeftCol) * CharWidth,
                   (FCaret.Y - TopLine) * LineHeight );
      ShowCaret( Control.Handle );
   end
   else
      HideCaret( Control.Handle );
   AdjustHScroll;
end;

procedure THIHiLightMemo.SetLeftCol(const Value: Integer);
var   WasLeftCol: Integer;
begin
   WasLeftCol := FLeftCol;
   if WasLeftCol <> Value then begin
      FLeftCol := Value;
      if FLeftCol < 0 then
         FLeftCol := 0;
      if FLeftCol >= MaxLineWidth then
         FLeftCol := MaxLineWidth;
      if FLeftCol < 0 then
         FLeftCol := 0;
      ScrollWindowEx( Control.Handle, (-FLeftCol + WasLeftCol)*CharWidth,
         0, nil, nil, 0, nil, SW_INVALIDATE );
      Caret := Caret;
   end;
   //      :
   AdjustHScroll;
end;

procedure THIHiLightMemo.SetLines(Idx: Integer; const Value: String);
var   U: String;
begin
   Changing;
   while FLines.Count <= Idx do begin
      FLines.Add( '' );
      if not FUndoingRedoing then
         FUndoList.Add( 'A' );
   end;
   if FLines.Items[ Idx ] <> Value then begin
      if not FUndoingRedoing then begin
         U := 'R' + Int2Str( Idx ) + '=';
         if Copy( FUndoList.Last, 1, Length( U ) ) <> U then
            FUndoList.Add( U + FLines.Items[ Idx ] );
      end;
      FLines.Items[ Idx ] := Value;
   end;
   InvalidateLine( Idx );
   Changed;
end;

procedure THIHiLightMemo.SetOnScanToken(const Value: TOnScanToken);
begin
   FOnScanToken := Value;
   Control.Invalidate;
end;

procedure THIHiLightMemo.SetSel(const Pos1, Pos2, PosFrom: TPoint);
begin
   if (Pos1.Y > Pos2.Y) or (Pos1.Y = Pos2.Y) and (Pos1.X > Pos2.X) then begin
      SelBegin := Pos2;
      SelEnd := Pos1;
   end
   else begin
      SelBegin := Pos1;
      SelEnd := Pos2;
   end;
   SelFrom := PosFrom;
end;

procedure THIHiLightMemo.SetSelBegin(const Value: TPoint);
begin
   if (FSelBegin.X = Value.X) and (FSelBegin.Y = Value.Y) then Exit;
   InvalidateSelection; //   ,   
   FSelBegin := Value;
   if FSelBegin.Y < 0 then FSelBegin.Y := 0; 
   if FSelBegin.X < 0 then FSelBegin.X := 0;
   if (FSelEnd.Y < FSelBegin.Y) or 
      (FSelEnd.Y = FSelBegin.Y) and
      (FSelEnd.X < FSelBegin.X) then
      FSelEnd := FSelBegin;
   InvalidateSelection; //   
   if Assigned( Control.OnSelChange ) then
      Control.OnSelChange( Control );
end;

procedure THIHiLightMemo.SetSelection(const Value: String);
var   S1, L: String;
      SL: PStrList;
      y: Integer;
begin
   //  ,    ,    
   Changing;
   DeleteSelection;
   if Value <> '' then begin
      SL := NewStrList;
TRY
   SL.Text := Value;
   if (Value <> '') and (Value[ Length( Value ) ] in [ #13, #10 ]) then
      SL.Add( '' );
      L := Lines[ Caret.Y ];
      if Length( L ) < Caret.X then
         L := L + StrRepeat( ' ', Caret.X - Length( L ) );
      S1 := ''; //     
      if SL.Count = 1 then begin
         L := Copy( L, 1, Caret.X ) + SL.Items[ 0 ] + CopyEnd( L, Caret.X+1 );
         FCaret.x := FCaret.x+Length( SL.Items[ 0 ] );
         Lines[ Caret.Y ] := L;
      end
      else begin
         S1 := CopyEnd( L, Caret.X+1 );
         L := Copy( L, 1, Caret.X ) + SL.Items[ 0 ];
         //S1 := SL.Items[ Count-1 ];
         Lines[ Caret.Y ] := L;
         FixBookmarks( Caret, Length( SL.Items[ 0 ] ) - Length( S1 ), SL.Count-1 );
         for y := 1 to SL.Count-2 do begin
            L := SL.Items[ y ];
            InsertLine( Caret.Y+y, L, FALSE );
         end;
         y := SL.Count-1;
         L := S1; 
         InsertLine( Caret.Y+y, SL.Last + L, FALSE );
         Caret := MakePoint( Length( SL.Last ), Caret.y+y );
      end;
FINALLY
   SL.Free;
END;
   end;
   AdjustHScroll;
   AdjustVScroll;
   SetSel( Caret, Caret, Caret );
   CaretToView;
   Changed;
end;

procedure THIHiLightMemo.SetSelEnd(const Value: TPoint);
begin
   if (FSelEnd.X = Value.X) and (FSelEnd.Y = Value.Y) then Exit;
   InvalidateSelection; //   ,   
   FSelEnd := Value;
   if FSelEnd.Y < 0 then FSelEnd.Y := 0; 
   if FSelEnd.X < 0 then FSelEnd.X := 0;   
   if (FSelBegin.Y > FSelEnd.Y) or
      (FSelBegin.Y = FSelEnd.Y) and
      (FSelBegin.X > FSelEnd.X) then
      FSelBegin := FSelEnd;
   InvalidateSelection; //   
   if Assigned( Control.OnSelChange ) then
      Control.OnSelChange( Control );
end;

procedure THIHiLightMemo.SetText(const Value: String);
begin
   FLines.Text := Value;
   TopLine := TopLine;
   Control.Invalidate;
   AutoCompletionHide;
   FastFillDictionary;
end;

procedure THIHiLightMemo.SetDrawRightMargin(const Value: Boolean);
begin
   FRightMargin := Value;
end;

procedure THIHiLightMemo.SetTopLine(const Value: Integer);
var   WasTopLine: Integer;
begin
   WasTopLine := FTopLine;
   FTopLine := Value;
   if FTopLine < 0 then
      FTopLine := 0;
   if FTopLine >= Count then 
      FTopLine := Count;
   if Count = 0 then begin
      Exit;
   end;
   if FTopLine + LinesPerPage >= Count then
      FTopLine := Count - LinesPerPage;
   if FTopLine < 0 then
      FTopLine := 0;
   if WasTopLine <> FTopLine then begin
      ScrollWindowEx( Control.Handle, 0, (WasTopLine - FTopLine)*LineHeight,
                      nil, nil, 0, nil, SW_INVALIDATE );
      Caret := Caret;
   end;
   //      :
   AdjustVScroll;
end;

procedure THIHiLightMemo.Undo;
begin
   if not CanUndo then Exit;
   DoUndoRedo( FUndoList, FRedoList );
end;

function THIHiLightMemo.WordAtPos(const AtPos: TPoint): String;
var   FromPos: TPoint;
      S:       String;
      i:       Integer;
begin
   Result := '';
   if (AtPos.Y < 0) or (AtPos.X < 0) then exit;  
   if AtPos.Y >= Count then Exit;
   FromPos := WordAtPosStart( AtPos );
   S := Lines[ FromPos.Y ];
   i := FromPos.X;
   while TRUE do begin
      inc( i );
      if i >= Length( S ) then break;
      if not IsLetterDigit( S[ i+1 ] ) then break;
   end;
   Result := Trim( Copy( S, FromPos.X+1, i-FromPos.X ) );
end;

function THIHiLightMemo.WordAtPosStart(const AtPos: TPoint): TPoint;
var   S: String;
      i: Integer;
begin
   Result := AtPos;
   if (AtPos.Y < 0) or (AtPos.X < 0) then exit; 
   if AtPos.Y >= Count then Exit;
   S := Lines[ AtPos.Y ];
   if (AtPos.X < Length( S )) and
       IsLetterDigit( S[ AtPos.X+1 ] ) then
      i := AtPos.X
   else if (AtPos.X-1 > 0) and
           (AtPos.X-1 < Length( S )) and
            IsLetterDigit( S[ AtPos.X ] ) then
      i := AtPos.X-1
   else Exit;
   while i > -1 do begin
      Dec( i );
      if (i < 0) or not IsLetterDigit( S[ i+1 ] ) then begin
         Result.X := i+1;
         break;
      end;
   end;
end;

function CheckKeyMask(State:byte; Code:byte):boolean;
begin
   case State of
      1: Result := GetKeyState(Code) < 0;
      else Result := true;
   end;
end;

initialization
   InitUpper;

end.